!+
! Module: epevents_mod
!
! Authors:  Matthias Ehle    - XMM-SSC-MPE
!           Michael Freyberg - XMM-SSC-MPE (mjf@mpe.mpg.de)
!           Wolfgang Pietsch - XMM-SSC-MPE
!
! Description: Converts an EPIC PN raw events file into a calibrated event
! list corrected for gain variations of the preamplifier channels and
! charge transfer inefficiency (CTI).
! It flags trailing events and performs a pattern analysis to identify and
! recombine split events. It takes into account bad pixel information from
! the badpix extension of the input file and discarded lines from the
! exposure extension.
!
! Comments: simple test call:
!      epevents eventset=rawevents01.dat outset=events01.dat
!
!-

module epevents_mod

  use eventAttributes
  use cal
  use dal
  use param
  use errorHandling
  use sas_epn_gaincticorrect_mod

  implicit none

  private

  ! Global variables

  character(len=10), parameter :: CodeVersion = '6.32'
  character(len=1024):: outfilename, eventmapset

  integer, parameter :: nx_ccd = EPN_MAX_X   ,& ! CCD-columns = 64
                        ny_ccd = EPN_MAX_Y   ,& ! CCD-rows   = 200
                        nphotonband = 4         ! number of photon map bands

  type(DataSetT) :: infile, outfile, outmip
  type(ArrayT)   :: b1, b2, b3, b4
  type(TableT)   :: events, badpix , exposure
  type(ColumnT)  :: frame, detX, detY, qualflag, time, &
                    badrawX, badrawY, badPixType, yextent, badflag, &
                    framelim, ndsclin1, ndsclin2, ndsclin3, ndsclin4, &
                    pat_typ, pat_ind, pat_ori, pat_seq, &
                    pat_id, pattern, pi, pha_gain, pha_gainFast, pha_ctiRaw

  integer :: maxnevf, &      ! maximum number of events in frames
             maxpat, &       ! max. pattern number per frame (maxnef/2)
             maxeqv, &       ! max nr of event classes (maxnef/3)
             maxcls, &       ! max nr of entries in event class (maxnef)
             maxsel, &       ! max selectable image parts (maxnef)
             nxdim, nydim, & ! window size
             nx0, ny0, &     ! bottom left corner of window
             nevent, &       ! number of events (rows in events extension of
                             ! input file)
             ifrm, &         ! value of frame processed
             nfrm, &         ! number of frames (rows in exposure extension of
                             ! input file, including empty frames)
             ntrail, &       ! number of trailing events
             n1, n2, n3, n4,&! number of singles, doubles, triples, quadruples
             nmip, &         ! number of MIPs
             n0, nrest, &    ! not recognized events, not recognized patterns
             nmax, &         ! maximal pattern size
             npat, &         ! number of detected patterns
             ccd, ccdno, &   ! overall CCD number (1-12)
             instmode, &     ! instrument data readout mode
             srcpos = 0, &   ! source position for TIMING and BURST mode
             eamipsel,  &    ! additional columns flagged for ONBOARD mip mode
             ixst,ixen, &    ! window corners in x
             iyst,iyen, &    ! window corners in y
             ixste,ixene, &  ! extended window corners in x
             iyste,iyene, &  ! extended window corners in y
             ievf, jevf, &   ! OOTE processing
             ignore_rest     ! skip rest of EXPOSURE extension

  integer    x2(1:2), y2(1:2), a2(1:2), p2, m21, m22, pp

  real(kind=double) :: d_srcpos = 0.0 ! source position for TI + BU mode

  logical :: lflag1, lmipm(4), linstmim, linstmti, linstmbu, lburst
  logical :: li2mapold(1:64)  ! Flags for changed i2mapf columns
  logical :: li2mpcold(0:65)  ! Flags for changed i2mpcf columns
!
! enable or disable certain features at compile time:
!
  logical(kind=bool), parameter :: enable_tempcorr = .true.
  logical(kind=bool), parameter :: enable_oote = .true.
  logical(kind=bool), parameter :: enable_out_of_fov = .true.
  logical(kind=bool), parameter :: enable_update_flag = .false.  ! for OoT events
  logical(kind=bool), parameter :: enable_scr_48 = .true.
! logical(kind=bool), parameter :: enable_scr_88 = .false.
! logical(kind=bool), parameter :: enable_scr_88 = .true.
  logical(kind=bool), parameter :: enable_timing_gain = .true.
  logical(kind=bool), parameter :: enable_burst_gain = .true.
!
! CTI calibrations have been performed with ODF-related frame numbers
! this is implemented here, just in case someone may want to change this:
!
  logical            :: frameCTI = .true.     ! use TIME-related frame numbers
! logical            :: frameCTI = .false.    !   instead of ODF frame numbers
  logical            :: enable_debug_TI_BU = .false.
  logical            :: enable_scr_88 = .true.


  integer(kind=Int32), dimension(1) :: dims

  real(kind=double),   dimension(:), pointer :: timeCol
  real(kind=single),   dimension(:), pointer :: ycolwrk
  integer(kind=Int32), dimension(:,:), pointer :: &
                                     b1_arrayP, b2_arrayP, b3_arrayP, b4_arrayP
  integer(kind=Int32), dimension(:), pointer :: frameCol, qualflagCol
  integer(kind=Int16), dimension(:), pointer :: rawXCol, rawYCol, &
                                                detXCol, detYCol, &
                                                phaCol, piCol, pha_ctiCol, &
                                                pha_gainCol, pha_gainFastCol, &
                                                pha_ctiRawCol, &
                                                pat_typCol, pat_indCol, &
                                                pat_oriCol, pat_idCol, &
                                                badrawXCol, badrawYCol, &
                                                badPixTypeCol, &
                                                yextentCol, badflagCol, &
                                                ndsclin1Col, ndsclin2Col, &
                                                ndsclin3Col, ndsclin4Col, &
                                                tmpRawYCol
  integer(kind=Int8),  dimension(:), pointer :: patternCol, pat_seqCol

  integer(kind=Int32), dimension(2)          :: imageDim
  integer(kind=Int32)                        :: b1_array(1:nx_ccd,1:ny_ccd), &
                                                b2_array(1:nx_ccd,1:ny_ccd), &
                                                b3_array(1:nx_ccd,1:ny_ccd), &
                                                b4_array(1:nx_ccd,1:ny_ccd)

! real(kind=single),dimension(:), allocatable    :: eTot

  integer(kind=Int32),dimension(:), allocatable  :: ind

  integer,dimension(:,:),allocatable             :: i2map, &
                                       ! bad pixel map
                                                    i2mapf, & 
                                       ! bad pixel map and disc line map 
                                                    i2mpc, &
                                       ! "convolved" bad pixel map
                                                    i2mpcf, &
                             ! "convolved" bad pixel and discarded line map
                                                    i4map , indmap
                             !  maps for split to make them allocatable

  real(kind=single),dimension(:,:), allocatable  :: r4map
  integer(kind=Int16),dimension(:,:),allocatable :: ndsclin

  public epevents

contains 

!---------------------------------------------------------------------
subroutine epevents
!---------------------------------------------------------------------

  character(len=80)                    :: map_title,fileMap
  character(len=8)                     :: eventMapExtension
  character(len=1),dimension(:)        :: mapPatternType(1:nphotonband)

  real(kind=double)                    :: intemperature, &
                                          scale_x, scale_y, &
                                          t0, frmtime, ymin,ymax, &
                                          tstart, tstop

  real(kind=double) :: corrpar(12) = (/ 0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0. /)
  real(kind=double) :: corrpar_in(5) = (/2.56d0,1.15d0,-0.26d0,0.10d0,120.0d0/)

  real(kind=single)                    :: lothresh(1:nphotonband), &
                                          hithresh(1:nphotonband)

  real(kind=single)                    :: amip,thp, temperature_gain
  real(kind=single),parameter          :: temp_gain_0080(0:3) = &
                                                  (/ 19.21,21.03,20.29,19.24 /)

  integer(kind=Int32)                  :: gainctiaccuracy

  integer(kind=Int16)                  :: ithp,itrail,iymin,iymax,ib
  integer(kind=Int16)                  :: quadrant, ccdid ! [0..3][0..2]

  logical,dimension(:), allocatable    :: lqual

  real(kind=single),  dimension(:),pointer     :: eOut, eIn

! integer(kind=int16),dimension(:),pointer     :: ieIn
  integer(kind=int16),dimension(:,:),pointer   :: FOVmap

  real(kind=single),  dimension(:),allocatable :: ampl    ! ,eInReal
  integer(kind=Int32),dimension(:),allocatable :: iqual, iqual0, jqual
  integer(kind=Int32),dimension(:),allocatable :: iframe
  integer(kind=int16),dimension(:),allocatable :: iyoote, ixev, iyev, &
                                                  ipat1,ipat2,ipat3,ipat4
  integer(kind=int16),dimension(:),allocatable,target :: tmpRawY
  integer(kind=int8) ,dimension(:),allocatable :: ipattern
  integer(kind=int16)                          :: ipat2_err

  integer                                      :: nevf, & 
                                             ! number of events in frame
                                                  irow,irowen, irow_old, &
                                             ! start, end row of frame (0,..)
                                                  k1, k2, k3, k4, &
                                                  krest, kmax, kmip, iamip, &
                                                  i4buf(3,3), i4dist(3,3), &
                                                  i,j,ij,jj,kj,kpind,nrowen

  logical :: prandom, &       ! trigger for randomization of position
             erandom, &       ! trigger for randomization of energy
             do_oote, &       ! trigger for out-of-time events computation
             do_tempcorr, &   ! trigger for additional PI correction
             do_ccdoffs, &    ! trigger for additional PI correction
             do_lt_cti,     & ! trigger for additional PI correction
             do_lt_cti_soft,& ! trigger for additional PI correction
             do_lt_cti_y,   & ! trigger for additional PI correction
             do_gain_TI,    & ! trigger for additional PI correction
             do_gain_BU,    & ! trigger for additional PI correction
             withphotonmap, & ! trigger for photon map output
             twidth , &       ! trigger for non-standard bin width
             lrecombraw       ! trigger for recombination of raw amplitudes
                              ! (as needed for CAL_pnGainCorrection)

! logical(kind=bool) :: ignore_rest = .false. ! skip rest of EXPOSURE extension
  logical(kind=bool) :: MIP_ext = .false.     ! withphotonmap
  logical(kind=bool) :: hasFRAMETIM = .true.  ! has FRAMETIM attribute


  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  ! ==================
  ! Start the programm
  ! ==================

  write(errstr,*) 'EPEVENTS ',trim(CodeVersion),' is running...'
  call message(AppMsg,Sparse)

  ! -----------------------
  ! Read program parameters
  ! -----------------------

  call getInParamValues(ithp,prandom,erandom,intemperature, &
                        gainctiaccuracy, twidth, do_oote, do_tempcorr, &
                        do_ccdoffs,do_lt_cti_soft,do_lt_cti_y, &
                        do_lt_cti, frameCTI, withphotonmap, &
                        nphotonband, lothresh, hithresh, mapPatternType, &
                        do_gain_TI, do_gain_BU, corrpar_in)

  thp = real(ithp)
  MIP_ext = withphotonmap

  ! -------------------
  ! Prepare data access
  ! -------------------

  call declareDal(ithp,prandom,erandom,intemperature,gainctiaccuracy, &
                  twidth,do_oote,quadrant,ccdId,withphotonmap)
  ccdNo = 3*quadrant+ccdId+1

  ! ----------------------------------------
  ! Get parameters from extension attributes
  ! ----------------------------------------



  write(errstr,*) 'Out_of_FOV: [1] IMAGING = ',linstmim,'  ENABLE = ', enable_out_of_fov
  call message(AppMsg,Noisy)
  call getAttrParamValues(nxdim,nydim,nx0,ny0,srcpos,iamip,eamipsel,&
                          tstart,tstop)
  write(errstr,*) 'Out_of_FOV: [1] IMAGING = ',linstmim,'  ENABLE = ', enable_out_of_fov
  call message(AppMsg,Noisy)

  if (.not.linstmim .and. frameCTI) enable_debug_TI_BU = .true.

! print*,'SRCPOS = ',srcpos

  ! limiting amplitude for MIPs: 

  amip = real(iamip) + 0.5
! ignore_rest = .false.
  ignore_rest = 0

  b1_array(1:nx_ccd,1:ny_ccd) = 0
  b2_array(1:nx_ccd,1:ny_ccd) = 0
  b3_array(1:nx_ccd,1:ny_ccd) = 0
  b4_array(1:nx_ccd,1:ny_ccd) = 0

  if (nevent > 0) then
    ! allocate i2map and i2mpc for badpix handling and
    ! allocate maps for discarded line handling (with window size)
    allocate(i2map(ixst:ixen,iyst:iyen))
    allocate(i2mpc(ixste:ixene,iyste:iyene))
    allocate(i2mapf(ixst:ixen,iyst:iyen))
    allocate(i2mpcf(ixste:ixene,iyste:iyene))

    ! ----------------------
    ! Bad pixel map handling
    ! ----------------------

    call bpxmap

    ! ----------------------------------------------
    ! Determine frame boundaries and discarded lines
    ! ----------------------------------------------

    call readFrames(temperature_gain)

    ! ----------------------------------------------------
    ! Issue some messages in case of experimental settings
    ! ----------------------------------------------------

    if (do_tempcorr) then
      if (temperature_gain < -273.2) then
        temperature_gain = temp_gain_0080(quadrant)
        write(errstr,'(a,a,f8.3,a)') &
          'No meaningful temperature for gain-temperature correction found. ',&
          'Will use in the following: ',temperature_gain,' [deg C]'
        call warning("noGainTemperature")
      endif
      write(errstr,'(a,f8.3,a)') 'Temperature for gain-temperature correction =',&
                                  temperature_gain,' [deg C]'
      call warning('ExperimentalCorrection', &
           'The use of temperature corrections is '// &
           'depreceated as it needs to be recalibrated.')
    endif

    if (do_ccdoffs) then
      call warning('ExperimentalCorrection', &
           'The use of CCD offsets is strongly depreceated.')
    endif

    if (do_lt_cti_soft .AND. do_lt_cti) then
      call warning('ExperimentalCorrection', &
           'The use of soft long-term CTI correction is strongly depreceated.')
    endif

!   if (do_lt_cti_y .AND. do_lt_cti) then
!     call warning('ExperimentalCorrection', &
!          'The use of Y-dependent long-term CTI correction is strongly '// &
!          'depreceated as it needs to be recalibrated.')
!   endif


    ! --------------
    ! Initialisation
    ! --------------
    call CAL_setState(EPN, ccdChipId = ccd, ccdModeId = instmode, &
                      accuracyLevel  = gainctiaccuracy, &
                      randomize = .false., &
                      temperature = intemperature)

    ! initialize the statistical variables:

    n1 = 0
    n2 = 0
    n3 = 0
    n4 = 0
    nrest = 0
    nmax = 0
    nmip = 0
    ntrail = 0
    maxnevf = 0
    i4dist = 0
    j = 0

    ! allocate maps for split handling and initialize
    allocate(i4map(ixste:ixene,iyste:iyene))
    i4map = 0
    if (verbosity(AppMsg) >= Noisy) then
      allocate(r4map(ixste:ixene,iyste:iyene))
      r4map = 0.0
    endif

    allocate(indmap(ixst:ixen,iyst:iyen))

    ! --------------------------
    ! Read table columns, 
    ! Start loop over all frames
    ! --------------------------

    nrowen = nevent - 1

    irow_old = -1

    iymin = MINVAL(rawYCol(0:nrowen))
    iymax = MAXVAL(rawYCol(0:nrowen))
    ymin  = iymin-0.4999999d0
    ymax  = iymax+0.4999999d0
!   print*,'Ymin = ',IYMIN,'   Ymax = ',IYMAX,'   Wmin = ',IYST,'   Wmax = ',IYEN

    do ifrm = 0, nfrm - 1
!
! ind(ifrm) refers to previous frame as numbering is [1..nfrm]
!
      irow = ind(ifrm)
      nevf = ind(ifrm + 1) - ind(ifrm)

!------------------------------------------------------------------------------
! next few lines for extended EXPOSURE extension (empty frames)
!
      if (irow == -1) then
        call message(AppMsg,Sparse, 'May now give FATAL error: FRAMELIM = -1')
        write(errstr,*) 'ifrm, irow, irow_old = ',ifrm, irow, irow_old
        call message(AppMsg,Sparse)
      endif
      irow_old = irow
!------------------------------------------------------------------------------

      if (nevf == 0) then
!       if (verbosity(AppMsg) > Noisy) then
!         write(errstr,*) 'Empty frame at IFRM = ',ifrm,' with IROW = ',irow
!         call message(AppMsg,Sparse)
!       endif
        cycle
      endif

      irowen = irow + nevf - 1

!     if (verbosity(AppMsg) >= Noisy) then
!       write(errstr,"(a,4i8)")" working on ifrm,irow,irowen,nevf", &
!             ifrm,irow,irowen,nevf 
!       call message(AppMsg,Noisy)
!     endif

      maxnevf = max(maxnevf,nevf) 
      maxpat = nevf / 2 + 1
!     maxeqv = nevf / 2 + 1
      maxeqv = nevf                  ! MJF@2002-01-14
      maxcls = nevf
      maxsel = nevf

      ! allocate array dimensions
      allocate(ixev(1:nevf))
      allocate(iyev(1:nevf))
      allocate(ampl(1:nevf))
      allocate(iqual(1:nevf))
      allocate(jqual(1:nevf))        ! quality of patterns alone
      allocate(lqual(1:nevf))
      allocate(ipat1(1:nevf))
      allocate(ipat2(1:nevf))
      allocate(ipat3(1:nevf))
      allocate(ipat4(1:nevf))

      ixev(1:nevf) = rawXCol(irow:irowen)
      iyev(1:nevf) = rawYCol(irow:irowen)
      ampl(1:nevf) = real(phaCol(irow:irowen))
      iqual(1:nevf) = qualflagCol(irow:irowen)
      jqual(1:nevf) = 0
      lqual(1:nevf) = (iqual(1:nevf) <= 24)     !! should be generalized !!

!     do i = 1,nevf
!       if (iqual(i) > 0) print*,'I,X,Y,IQUAL = ',i,ixev(i),iyev(i),iqual(i)
!     enddo
!     lqual(1:nevf) = (iqual(1:nevf) == 0)

      if (do_oote) then
        allocate(iyoote(1:nevf))
        allocate(ycolwrk(1:nevf))
        allocate(iqual0(1:nevf))
        iqual0(1:nevf) = iqual(1:nevf)
      endif

      if (nfrm > 10) then
        if (mod(ifrm,nint(nfrm/10.)) == 0) then
          write(errstr,"(3x,a,i9,4x,a,i9,3x,f5.1,a,a1)")      &
               "CCD frame:",ifrm+1,"event:",irow+1,           &
                real(irow+1)/real(nevent)*100.0,"% completion"
          call message(AppMsg,Verbose)
        endif
      else
        write(errstr,"(3x,a,i9,4x,a,i9,3x,f5.1,a,a1)")        &
               "CCD frame:",ifrm+1,"event:",irow+1,           &
                real(irow+1)/real(nevent)*100.0,"% completion"
        call message(AppMsg,Verbose)
      endif    

      ! ---------------------------------------------- 
      ! Determine i2mpcf from discarded lines in frame
      ! ----------------------------------------------

      if (.not.lmipm(1)) call discLines(ifrm)

      ! --------------------
      ! Flag trailing events
      ! --------------------

      ! only if more then 1 event and not timing mode
      if (nevf > 1 .and. .not.linstmti) then 
        call CAL_setState(EPN,accuracyLevel  = ACCURACY_HIGH) 
        call trailEvent(nevf,ixev,iyev,ampl,iqual,lqual,thp,itrail)
        ntrail = ntrail + itrail
        qualflagCol(irow:irowen) = iqual(1:nevf)
        call CAL_setState(EPN,accuracyLevel  = gainctiaccuracy) 
      endif

      ! ------------------------
      ! Perform pattern analysis
      ! ------------------------

!     print*,'split ...'
      call split(nevf,ixev,iyev,ampl,iqual, &
                      lqual,amip, &
                      ipat1,ipat2,ipat3,ipat4,i4buf,k1,k2,k3,k4,     &
                      krest,kmax,kmip,do_oote) 

      ! --------------------------------
      ! Collect the pattern informations
      ! --------------------------------

      call collectPattern(irow,irowen, nevf,ipat1,ipat2,ipat3,ipat4,iqual, &
!                         jqual, &
                          lqual,i4buf,k1,k2,k3,k4,krest,kmax,kmip) 

! The Out-of-Time Events section has been written by Alexis Finoguenov with
! minor modifications by Michael Freyberg

      if (do_oote) then

! before the ipat2 array is freed we need to make OOTE simulation. Alexis.
! Randomize Y-column. The procedure sounds a bit too complicated, however.

        call CAL_setState(EPN,randomize = .true.)
!       call CAL_setRandom(+11.5d0,+200.833333d0) ! is incorrect
        call CAL_setRandom(YMIN,YMAX)    ! these are correct and general values
        iyoote(1:nevf) = 0
        call CAL_integerToReal(iyoote,ycolwrk)
        call CAL_setState(EPN,randomize = .false.)
        iyoote(1:nevf) = nint(ycolwrk(1:nevf))

! Fix the split events to be in the same relative position to the main event

        do ievf = 1,nevf
          if (ipat2(ievf) < 0) then
            do jevf = 1,nevf
              if ( (-ipat2(ievf)) .eq. ipat2(jevf)) then
                iyoote(ievf) = iyoote(jevf) + iyev(ievf) - iyev(jevf)
              endif
            enddo
          endif
        enddo

! check if patterns are moved outside the CCD window:
! well, double events with one partner simulated at Y=200 may get the other 
! partner at Y=201, ok, for this rare case we have to cheat a bit in this
! simple model and to shift the whole pattern by 1 pixel to the CAMEX

        do ievf = 1,nevf
          if (iyoote(ievf) > iyen) then
            i = irow-1+ievf
!           print*,'Y-ERR: ',i,rawXCol(i),iyoote(ievf),NINT(ampl(ievf))
            ipat2_err = abs(ipat2(ievf))
            do jevf = 1,nevf          
              if (abs(ipat2(jevf)) .eq. ipat2_err) then
                iyoote(jevf) = iyoote(jevf)-1
                i = irow-1+jevf
!               print*,'Yjerr: ',i,rawXCol(i),iyoote(jevf),NINT(ampl(jevf))
              endif
            enddo
          endif
        enddo
        do ievf = 1,nevf
          if (iyoote(ievf) < iyst) then
            i = irow-1+ievf
!           print*,'Y-ERR: ',i,rawXCol(i),iyoote(ievf),NINT(ampl(ievf))
            ipat2_err = abs(ipat2(ievf))
            do jevf = 1,nevf          
              if (abs(ipat2(jevf)) .eq. ipat2_err) then
                iyoote(jevf) = iyoote(jevf)+1
                i = irow-1+jevf
!               print*,'Yjerr: ',i,rawXCol(i),iyoote(jevf),NINT(ampl(jevf))
              endif
            enddo
          endif
        enddo

! Fill out the rawYCol to complete the action. This frame has now only OOTE.

        rawYCol(irow:irowen) = iyoote(1:nevf)

        if (enable_update_flag) then
          iyev(1:nevf) = iyoote(1:nevf)

      ! ------------------------------------------------------------
      ! Re-Perform pattern analysis because patterns are now shifted
      ! ------------------------------------------------------------

          call split(nevf,ixev,iyev,ampl,iqual0, &
                      lqual,amip, &
                      ipat1,ipat2,ipat3,ipat4,i4buf,k1,k2,k3,k4,     &
                      krest,kmax,kmip,do_oote) 

      ! --------------------------------
      ! Collect the pattern informations
      ! --------------------------------

          call collectPattern(irow,irowen, nevf,ipat1,ipat2,ipat3,ipat4,iqual0,&
!                         jqual,
                          lqual,i4buf,k1,k2,k3,k4,krest,kmax,kmip) 

          iqual(1:nevf) = iqual0(1:nevf)
        endif ! update_flag

! End of OOTE part. Only memory deallocation is made below this line. Alexis.

      endif ! do_oote

      ! free the memory:
      deallocate(ixev)
      deallocate(iyev)
      if (do_oote) then
        deallocate(iyoote)
        deallocate(iqual0)
        deallocate(ycolwrk)
      endif
      deallocate(ampl)
      deallocate(iqual)
      deallocate(jqual)
      deallocate(lqual)
      deallocate(ipat1)
      deallocate(ipat2)
      deallocate(ipat3)
      deallocate(ipat4)

      ! ------------------------
      ! End loop over all frames
      ! ------------------------
    enddo

    ! deallocate badpix and discarded line and split maps
    deallocate(i2map)
    deallocate(i2mpc)
    deallocate(i2mapf)
    deallocate(i2mpcf)
    deallocate(i4map)
    if (verbosity(AppMsg) >= Noisy) deallocate(r4map)
    deallocate(indmap)
    deallocate(ndsclin)



    ! -------------------------------
    ! Perform Gain and CTI correction 
    ! -------------------------------
    if (verbosity(AppMsg) >= Verbose) &
       call message(AppMsg,Verbose,'performing GAIN correction ...')

    allocate(eIn(0:nrowen))
    allocate(eOut(0:nrowen))

    call CAL_setState(EPN,randomize = .false.)
    if (erandom) then
      call CAL_setState(EPN,randomize = erandom)
      call CAL_setRandom(-0.5d0,+0.5d0)
    endif
    call CAL_integerToReal(phaCol,eIn)
    call CAL_setState(EPN,randomize = .false.)

    if (linstmti .or. linstmbu) then
      d_srcpos = srcpos
      call CAL_setState(EPN, auxiliaryParameter = d_srcpos)
    endif

!   eIn(0:nrowen) = real(phaCol(0:nrowen))
    eOut(0:nrowen) = 0.0

    call CAL_pnGainCorrect(rawXCol, rawYCol, eIn, eOut)
    if ((linstmti.or.linstmbu).and.(enable_debug_TI_BU)) then
      pha_gain  = addColumn(events,"PHA_GAIN",Integer16,"channel",dims, &
             comment="GAIN-corrected Event Energy")
      call setAttribute(pha_gain, "TLMIN",-32767, comment="Minimum for PHA_GAIN")
      call setAttribute(pha_gain, "TLMAX",32767, comment="Maximum for PHA_GAIN")
      call setNullValue(pha_gain,0)
      pha_gainCol  => int16Data(pha_gain)
      pha_gainCol(0:nevent-1) = nint(eOut(0:nevent-1))
      call message(AppMsg,Sparse,' ... created PHA_GAIN column')
    endif
    if ((linstmti.or.linstmbu).and.(enable_debug_TI_BU)) then
      pha_gainFast  = addColumn(events,"PHA_GAINF",Integer16,"channel",dims, &
             comment="GAIN-corrected Event Energy")
      call setAttribute(pha_gainFast, "TLMIN",-32767, comment="Minimum for PHA_GAINF")
      call setAttribute(pha_gainFast, "TLMAX",32767, comment="Maximum for PHA_GAINF")
      call setNullValue(pha_gainFast,0)
      pha_gainFastCol  => int16Data(pha_gainFast)
      call message(AppMsg,Sparse,' ... created PHA_GAINF column')
    endif

    if (linstmti .and. do_gain_ti) then
      if (verbosity(AppMsg) >= Verbose) &
        call message(AppMsg,Sparse,'performing TI gain fine-tuning ...')
      call SAS_pnGainTimingCorrect(eOut,nevent)
    elseif (linstmbu .and. do_gain_bu) then
      if (verbosity(AppMsg) >= Verbose) &
        call message(AppMsg,Sparse,'performing BU gain fine-tuning ...')
      call SAS_pnGainBurstCorrect(eOut,nevent)
    endif

    if ((linstmti.or.linstmbu).and.(enable_debug_TI_BU)) then
      pha_gainFastCol(0:nevent-1) = nint(eOut(0:nevent-1))
    endif

    if (verbosity(AppMsg) >= Verbose) &
      call message(AppMsg,Verbose,'performing CTI correction ...')

    hasFRAMETIM = hasAttribute(events,"FRAMETIM")
    allocate(iframe(0:nrowen))

    call setAttribute(events, "FRAMECTI",frameCTI, &
                      comment="use TIME frame numbers instead of ODF numbers")

    call setAttribute(events, "SCR_88",enable_scr_88, &
                      comment="use PATTERN coding as of SOC_SCR_88")

    if (frameCTI .and. hasFRAMETIM) then  ! use TIME instead of FRAME_ODF
      time = column(events,"TIME",READ)
      timeCol => real64Data(time)
      t0 = timeCol(0)
      frmtime = real64Attribute(events,"FRAMETIM")*1.D-3         ! [ms] -> [s]
      if ((.not.linstmim).and.(frameCTI)) then
        write(errstr,'(a,a,a,f10.6,a)') &
              'frameCTI = Y for non-IMAGING modes. ', &
              'This may result in spectral broadening at the lowest energies. ',&
              'Will use FRAMETIM = ',frmtime*1.d3, &
              ' [ms] to convert TIME to FRAME for CTI correction.'
        call warning('frameCTI')
      endif
      iframe(0:nrowen)   = NINT((timeCol(0:nrowen)-t0)/frmtime)
!------ DEBUG
!     frame = column(events,"FRAME",READ)
!     frameCol => int32Data(frame)
!     do j = 0,nevent-1
!       if (frameCol(j) .ge. 433 .and. frameCol(j) .le. 470) then
!         print*,timeCol(j),frameCol(j),iframe(j),rawXCol(j),rawYCol(j),phaCol(j)
!       endif
!     enddo
!     call release(frame)
!------ DEBUG
      call release(time)
    else                                    ! use FRAME_ODF
      frame = column(events,"FRAME",READ)
      frameCol => int32Data(frame)
      iframe(0:nrowen)   = frameCol
      call release(frame)
    endif
    if (frameCTI .and. .NOT. hasFRAMETIM) then  ! use TIME instead of FRAME_ODF
      call message(AppMsg,Sparse,'EVENTS file does not have FRAMETIM attribute')
    endif

    eIn(0:nrowen) = eOut(0:nrowen)
    call CAL_pnCtiCorrect(rawXCol, rawYCol, iframe, eIn, eOut)

    if ((linstmti.or.linstmbu).and.(enable_debug_TI_BU)) then
      pha_ctiRaw  = addColumn(events,"PHA_CTIR",Integer16,"channel",dims, &
             comment="CTI-corrected Event Energy")
      call setAttribute(pha_ctiRaw, "TLMIN",-32767, comment="Minimum for PHA_CTIR")
      call setAttribute(pha_ctiRaw, "TLMAX",32767, comment="Maximum for PHA_CTIR")
      call setNullValue(pha_ctiRaw,0)
      pha_ctiRawCol  => int16Data(pha_ctiRaw)
      pha_ctiRawCol(0:nevent-1) = nint(eOut(0:nevent-1))
      call message(AppMsg,Sparse,' ... created PHA_CTIR column')
    endif


!   eOut = eOut 

    ! adjustment Orsay/in-orbit + additional effects:

!   write(*,*) '-------------- call SAS_pnGainCtiCorrect() --------------'
    if (linstmti.or.linstmbu) then
      allocate(tmpRawY(0:nevent-1))
      tmpRawY(0:nevent-1) = srcpos
      tmpRawYCol => tmpRawY
      call SAS_pnGainCtiCorrect(eOut,tmpRawYCol,ccd,temperature_gain,tstart,tstop,&
                              do_tempcorr,do_ccdoffs,do_lt_cti_soft,&
                              do_lt_cti_y, do_lt_cti, corrpar, nevent,&
                              corrpar_in)
      deallocate(tmpRawY)
    else
      call SAS_pnGainCtiCorrect(eOut,rawYCol,ccd,temperature_gain,tstart,tstop,&
                              do_tempcorr,do_ccdoffs,do_lt_cti_soft,&
                              do_lt_cti_y, do_lt_cti, corrpar, nevent,&
                              corrpar_in)
    endif
    call CAL_realToInteger(eOut,pha_ctiCol)
    if (.not. twidth) deallocate(eOut)

    deallocate(eIn)
    deallocate(iframe)

    ! -------------------------------
    ! Recombine corrected amplitudes,
    ! loop over all frames
    ! -------------------------------
    lflag1=.true.  
    lrecombraw = .false. ! complete recombination 

    irow_old = -1

    do ifrm = 0, nfrm - 1  

      irow = ind(ifrm)
      nevf = ind(ifrm + 1) - ind(ifrm)

      if (nevf == 0) cycle
      irow_old = irow
      irowen = irow + nevf - 1

      maxpat = nevf / 2 + 1
      maxpat = max(maxpat,1)
      maxcls = nevf

      allocate(ampl(1:nevf))
      allocate(ipat1(1:nevf))
      allocate(ipat2(1:nevf))
      allocate(ipat3(1:nevf))
      allocate(iqual(1:nevf))
      allocate(lqual(1:nevf))
      allocate(ipattern(1:nevf))

      if (twidth) then
        ampl(1:nevf)   = eOut(irow:irowen) * 5.0        ! convert [adu] -> [eV]
      else
        ampl(1:nevf)   = pha_ctiCol(irow:irowen) * 5.0  ! convert [adu] -> [eV]
      endif
      ipat1(1:nevf)  = pat_typCol(irow:irowen)
      ipat2(1:nevf)  = pat_indCol(irow:irowen)
      ipat3(1:nevf)  = pat_oriCol(irow:irowen)
      ipattern(1:nevf)  = patternCol(irow:irowen)
      iqual(1:nevf)  = qualflagCol(irow:irowen)
      do i = 1, nevf
        lqual(i)  = .not.((iand(iqual(i),EVATT_IN_BAD_FRAME) > 0) .or. &
                          (iand(iqual(i),EVATT_MIP_ASSOCIATED) > 0) .or. &
                          (iand(iqual(i),EVATT_TRAILING) > 0))
      enddo
      call recombine(irow,irowen,nevf,ampl,ipat1,ipat2,iqual,lqual, &
                     lrecombraw,lflag1)

!
! convert the negative energy sums to positive ones
!
      do j = 1,nevf
        jj = irow-1+j
        if (piCol(jj) < 0 .and. ipat3(j) == 0 .and. abs(ipat1(j)) > 1) then
          piCol(jj) = -piCol(jj)
          kpind = ipat2(j)          ! PAT_IND of main event with negative sum
          do ij = 1,nevf            ! check all events in this frame with same
            if (ipat2(ij) == kpind) then  ! PAT_IND and reset flag information
              kj = irow-1+ij
              qualflagCol(kj) = ior(qualflagCol(kj),EVATT_INVALID_PATTERN)
            endif
          enddo
        endif
      enddo

!
! convert the "difficult-to-understand-without-reading-the-documentation"
! PATTERN values (like 128) to smaller numbers that may be easier to handle
! (like 0) [SOC-SCR-88]
!
      if (enable_scr_88) then

      do j = 1,nevf
        jj = irow-1+j
        pp = patternCol(jj)
        if (pp < 0) pp = pp+256
        if (pp == 128) then
!         write(errstr,*) '128: J = ',j,'  JJ = ',jj,'   -> 0'
!         call message(AppMsg,Sparse)
          patternCol(jj) = 0
        endif
        if (pp == 205) then
          p2 = 0
          x2(1:2) = 0
          y2(1:2) = 0
          a2(1:2) = 0
          kpind = ipat2(j)
          do ij = 1,nevf
            if (ipat2(ij) == kpind) then
              p2 = p2+1
              if (p2 > 2) then
                call warning("SOC-SCR-88","too many double events in pattern")
                exit
              endif
              kj = irow-1+ij
              x2(p2) = rawXCol(kj)
              y2(p2) = rawYCol(kj)
              a2(p2) = phaCol(kj)
            endif
          enddo
          if (a2(1) > a2(2)) then
            m21 = 1
            m22 = 2
          else
            m21 = 2
            m22 = 1
          endif
          if (x2(1) == x2(2)) then
            if (y2(m21) > y2(m22)) then
!             write(errstr,*) '205: J = ',j,'  JJ = ',jj,&
!                         '  x2 = ',x2(1),x2(2),&
!                         '  y2 = ',y2(1),y2(2),&
!                         '  a2 = ',a2(1),a2(2),'  -> 3'
!             call message(AppMsg,Sparse)
              patternCol(jj) = 3
            else
!             write(errstr,*) '205: J = ',j,'  JJ = ',jj,&
!                         '  x2 = ',x2(1),x2(2),&
!                         '  y2 = ',y2(1),y2(2),&
!                         '  a2 = ',a2(1),a2(2),'  -> 1'
!             call message(AppMsg,Sparse)
              patternCol(jj) = 1
            endif
          elseif (y2(1) == y2(2)) then
            if (x2(m21) > x2(m22)) then
!             write(errstr,*) '205: J = ',j,'  JJ = ',jj,&
!                         '  x2 = ',x2(1),x2(2),&
!                         '  y2 = ',y2(1),y2(2),&
!                         '  a2 = ',a2(1),a2(2),'  -> 4'
!             call message(AppMsg,Sparse)
              patternCol(jj) = 4
            else
!             write(errstr,*) '205: J = ',j,'  JJ = ',jj,&
!                         '  x2 = ',x2(1),x2(2),&
!                         '  y2 = ',y2(1),y2(2),&
!                         '  a2 = ',a2(1),a2(2),'  -> 2'
!             call message(AppMsg,Sparse)
              patternCol(jj) = 2
            endif
          endif
        endif
      enddo

      endif   ! enable_scr_88

      ! deallocate arrays
      deallocate(ampl)
      deallocate(ipat1)
      deallocate(ipat2)
      deallocate(ipat3)
      deallocate(ipattern)
      deallocate(iqual)
      deallocate(lqual)

      ! ------------------------
      ! End loop over all frames
      ! ------------------------
    enddo

    if (twidth) deallocate(eOut)

    ! deallocate ind
    deallocate(ind)

    ! ---------------------------------------------------
    ! Show results of trailing event and pattern analysis
    ! ---------------------------------------------------

    n0 = nevent - n1 - n2 * 2 - n3 * 3 - n4 * 4

    call message(AppMsg,Verbose,"Results of the pattern analysis")
    call message(AppMsg,Verbose,"-------------------------------")
    write(errstr,"(a,i8)") "number of analysed events:        ", nevent
    call message(AppMsg,Verbose)
    call message(AppMsg,Verbose,"")
    write(errstr,"(a,i8)") "number of trailing events:        ", ntrail
    call message(AppMsg,Verbose)
    write(errstr,"(a,i8)") "number of single events:          ", n1
    call message(AppMsg,Verbose)
    write(errstr,"(a,i8)") "number of double patterns:        ", n2
    call message(AppMsg,Verbose)
    write(errstr,"(a,i8)") "number of triple patterns:        ", n3
    call message(AppMsg,Verbose)
    write(errstr,"(a,i8)") "number of quadr. patterns:        ", n4
    call message(AppMsg,Verbose)
    write(errstr,"(a,i8)") "number of not recog. events:      ", n0
    call message(AppMsg,Verbose)
    write(errstr,"(a,i8)") "number of not recog. patterns:    ", nrest
    call message(AppMsg,Verbose)
    write(errstr,"(a,i8)") "maximum pattern size (pixel):     ", nmax
    call message(AppMsg,Verbose)
    write(errstr,"(a,i8)") "maximum number of events in frame:", maxnevf
    call message(AppMsg,Verbose)
    write(errstr,"(a,i8)") "number of MIPs found:             ", nmip
    call message(AppMsg,Verbose)
    call message(AppMsg,Verbose,"")
    write(errstr,"(a,i8)") "number of recognized photons:     ", n1+n2+n3+n4
    call message(AppMsg,Verbose)
  endif


  ! -----------------------------------------
  ! Convert from raw- to detector-coordinates 
  ! -----------------------------------------

  call CAL_setState(EPN, accuracyLevel = ACCURACY_HIGH)

  call raw2det(prandom,scale_x,scale_y)

  ! ---------------------------------------------------------
  ! determine and flag OUT_OF_FOV events: map is [0:63,0:199]
  ! ---------------------------------------------------------

  write(errstr,*) 'Out_of_FOV: [2] IMAGING = ',linstmim,'  ENABLE = ', enable_out_of_fov
  call message(AppMsg,Noisy)
  
  if (linstmim) then
    if (enable_out_of_fov) then
      call CAL_getFOVmap(FOVmap)
      do i = 0,nevent-1
        if (FOVmap(rawXCol(i)-1,rawYCol(i)-1) == 0) &
              qualflagCol(i) = ior(qualflagCol(i),EVATT_OUT_OF_FOV)
      enddo
      write(errstr,*) 'Out_of_FOV: determined via CAL_getFOVmap()'
      call message(AppMsg,Noisy)
    endif
  endif

  ! -------------------------------------------------
  ! drop redundant information of PI and FLAG columns
  ! -------------------------------------------------

  if (enable_scr_48) then
    where (piCol < 0 .AND. qualFlagCol > 0)
      piCol = -piCol
    end where
  endif

! if (enable_scr_88) then
!   where (piCol < 0 .AND. qualFlagCol = 0)
!     piCol = -piCol
!     qualflagCol = ior(qualflagCol,EVATT_INVALID_PATTERN)
!   end where
! endif

  ! ------------------------------------------
  ! Add to EXPOSURE extension (if it exists)
  ! ------------------------------------------

  call addExposure(ithp,prandom,erandom,intemperature,gainctiaccuracy, & 
                   scale_x,scale_y,twidth,corrpar,withphotonmap)


  if (withphotonmap) then
    imageDim(1) = nx_ccd
    imageDim(2) = ny_ccd

    outmip = dataSet(eventmapset,CREATE,USE_ENVIRONMENT)
    call addAttributes(attributable(outmip),attributable(outfile))

    write(eventMapExtension,'(a,i2.2)') 'PBAND1',CCDno
    b1 = addArray(outmip, eventMapExtension, Integer32, imageDim)
    write(eventMapExtension,'(a,i2.2)') 'PBAND2',CCDno
    b2 = addArray(outmip, eventMapExtension, Integer32, imageDim)
    write(eventMapExtension,'(a,i2.2)') 'PBAND3',CCDno
    b3 = addArray(outmip, eventMapExtension, Integer32, imageDim)
    write(eventMapExtension,'(a,i2.2)') 'PBAND4',CCDno
    b4 = addArray(outmip, eventMapExtension, Integer32, imageDim)

    b1_arrayP => int32Array2Data(b1)
    b2_arrayP => int32Array2Data(b2)
    b3_arrayP => int32Array2Data(b3)
    b4_arrayP => int32Array2Data(b4)

    call addAttributes(attributable(b1),attributable(exposure))
    call addAttributes(attributable(b2),attributable(exposure))
    call addAttributes(attributable(b3),attributable(exposure))
    call addAttributes(attributable(b4),attributable(exposure))

    ib = 1
    call setAttribute(b1, "MAP_ELO",lothresh(ib),"eV", &
       comment="photon map lower energy limit")
    call setAttribute(b1, "MAP_EHI",hithresh(ib),"eV", &
       comment="photon map upper energy limit")
    call setAttribute(b1, "MAP_PAT",mapPatternType(ib), &
       comment="photon map pattern type selection")
    call makeMapTitle(lothresh(ib),hithresh(ib),mapPatternType(ib),map_title)
    call setAttribute(b1, "MAPLATEX", map_title, &
               comment="Description of photon map")
    write(fileMap,'(a,2i2.2)') 'pmap',ib,CCDno
    call setAttribute(b1, "FILENAME", trim(fileMap), &
               comment="Identifier of photon map")

    ib = 2
    call setAttribute(b2, "MAP_ELO",lothresh(ib),"eV", &
       comment="photon map lower energy limit")
    call setAttribute(b2, "MAP_EHI",hithresh(ib),"eV", &
       comment="photon map upper energy limit")
    call setAttribute(b2, "MAP_PAT",mapPatternType(ib), &
       comment="photon map pattern type selection")
    call makeMapTitle(lothresh(ib),hithresh(ib),mapPatternType(ib),map_title)
    call setAttribute(b2, "MAPLATEX", map_title, &
               comment="Description of photon map")
    write(fileMap,'(a,2i2.2)') 'pmap',ib,CCDno
    call setAttribute(b2, "FILENAME", trim(fileMap), &
               comment="Identifier of photon map")

    ib = 3
    call setAttribute(b3, "MAP_ELO",lothresh(ib),"eV", &
       comment="photon map lower energy limit")
    call setAttribute(b3, "MAP_EHI",hithresh(ib),"eV", &
       comment="photon map upper energy limit")
    call setAttribute(b3, "MAP_PAT",mapPatternType(ib), &
       comment="photon map pattern type selection")
    call makeMapTitle(lothresh(ib),hithresh(ib),mapPatternType(ib),map_title)
    call setAttribute(b3, "MAPLATEX", map_title, &
               comment="Description of photon map")
    write(fileMap,'(a,2i2.2)') 'pmap',ib,CCDno
    call setAttribute(b3, "FILENAME", trim(fileMap), &
               comment="Identifier of photon map")

    ib = 4
    call setAttribute(b4, "MAP_ELO",lothresh(ib),"eV", &
       comment="photon map lower energy limit")
    call setAttribute(b4, "MAP_EHI",hithresh(ib),"eV", &
       comment="photon map upper energy limit")
    call setAttribute(b4, "MAP_PAT",mapPatternType(ib), &
       comment="photon map pattern type selection")
    call makeMapTitle(lothresh(ib),hithresh(ib),mapPatternType(ib),map_title)
    call setAttribute(b4, "MAPLATEX", map_title, &
               comment="Description of photon map")
    write(fileMap,'(a,2i2.2)') 'pmap',ib,CCDno
    call setAttribute(b4, "FILENAME", trim(fileMap), &
               comment="Identifier of photon map")

    call fillPhotonMap(b1_array,b2_array,b3_array,b4_array,nx_ccd,ny_ccd, &
                       lothresh,hithresh,mapPatternType,nphotonband)

    b1_arrayP = b1_array
    b2_arrayP = b2_array
    b3_arrayP = b3_array
    b4_arrayP = b4_array

    call release(outmip)
    call message(AppMsg,Verbose,'Releasing EXPOSURE extension ...')
    call release(exposure)
  endif

  call message(AppMsg,Verbose,'Releasing output file ...')
  call release(outfile)

  call message(AppMsg,Sparse,"EPEVENTS finished")

end subroutine epevents


! =========================
! Program parameter reading
! =========================
subroutine getInParamValues(ithp,prandom,erandom,intemperature, &
            gainctiaccuracy,twidth, do_oote, do_tempcorr, &
            do_ccdoffs, do_lt_cti_soft, do_lt_cti_y, &
            do_lt_cti, frameCTI, withphotonmap, &
            nphotonband, lothresh, hithresh, mapPatternType, &
            do_gain_TI, do_gain_BU, corrpar_in)

  integer            ,intent(in)    :: nphotonband
  real(kind=double)  ,intent(inout) :: corrpar_in(5)
  character(len=*)   ,intent(out)   :: mapPatternType(1:nphotonband)
  real(kind=double)  ,intent(out)   :: intemperature
  real(kind=single)  ,intent(out)   :: lothresh(1:nphotonband), &
                                       hithresh(1:nphotonband)
  integer(kind=Int32),intent(out)   :: gainctiaccuracy
  integer(kind=Int16),intent(out)   :: ithp
  logical            ,intent(out)   :: prandom, erandom, twidth, &
                                       do_oote, do_tempcorr, do_ccdoffs, &
                                       do_lt_cti_soft, do_lt_cti_y, &
                                       do_gain_TI, do_gain_BU, &
                                       do_lt_cti, withphotonmap, frameCTI
  character(len=1024)               :: mapPatternTypeString
  integer                           :: npar, i

  call message(AppMsg,Noisy,"... entering getInParamValues...")

  ! -----------------------------------------------------
  ! Get reemission threshold for trailing event rejection
  ! -----------------------------------------------------

  ithp = intParameter("reemissionthresh")

  write(errstr,'(a,i8)') 'Reemission threshold:   ', ithp
  call message(AppMsg,Noisy)

  ! ---------------------------------
  ! Get temperature of the camera [K]
  ! ---------------------------------

  intemperature = -9990.0    !  was:   realParameter("temperature")

! write(errstr,'(a,f6.1)') 'CCD temperature:        ',intemperature
! call message(AppMsg,Noisy)

  ! ---------------------------------
  ! Get accuraccy for gain correction
  ! ---------------------------------

  gainctiaccuracy = intParameter("gainctiaccuracy")

  write(errstr,*) 'Gain_cti_accuracy setting:      ',gainctiaccuracy
  call message(AppMsg,Noisy)

  ! -------------------------------------------
  ! Get special energy correction flags
  ! -------------------------------------------

  if (enable_tempcorr) then
    do_tempcorr    = booleanParameter("withtempcorrection")
    do_ccdoffs     = booleanParameter("withccdoffsets")
    do_lt_cti      = booleanParameter("withctilongterm")
    do_lt_cti_soft = booleanParameter("ctilongtermsoft")
    do_lt_cti_y    = booleanParameter("ctilongtermy")
  else
    do_tempcorr    = .false.
    do_ccdoffs     = .false.
    do_lt_cti      = .false.
    do_lt_cti_soft = .false.
    do_lt_cti_y    = .false.
  endif
  frameCTI         = booleanParameter("withframecti")

  ! -------------------------------------------
  ! Get special mode-gain flags
  ! -------------------------------------------

  if (enable_timing_gain) then
    do_gain_ti     = booleanParameter("withgaintiming")
  else
    do_gain_ti     = .false.
  endif
  if (enable_burst_gain) then
    do_gain_bu     = booleanParameter("withgainburst")
  else
    do_gain_bu     = .false.
  endif

  ! ------------------------------------------------------
  ! Randomization for calculation of detector coordinates?
  ! ------------------------------------------------------

  prandom = booleanParameter("randomizeposition")
  erandom = booleanParameter("randomizeenergy")
  twidth  = booleanParameter("testenergywidth")
  if (twidth) then
    write(errstr,*) 'You have requested a non-standard energy bin width ', &
                    'in the event amplitude correction. This parameter ', &
                    'is only to be used for calibration purposes. ', &
                    'This setting will yield ARTIFACTS ', &
                    'and as a consequence INCORRECT SCIENTIFIC RESULTS. ', &
                    'There is no support for this setting. ', &
                    'You have been warned. Good luck.'
    call warning('Twidth')
  endif

  ! -----------------------------------------------
  ! Do creation of out-of-time events file instead?
  ! -----------------------------------------------
  if (enable_oote) then
    do_oote = booleanParameter("withoutoftime")
  else
    do_oote = .false.
  endif

  withphotonmap = booleanParameter("withphotonmap")

  npar = parameterCount("lothresh")
  do i = 0, npar-1
    lothresh(i+1) = realParameter("lothresh",i)
  enddo
  npar = parameterCount("hithresh")
  do i = 0, npar-1
    hithresh(i+1) = realParameter("hithresh",i)
  enddo
  mapPatternTypeString = trim(stringParameter("mappatterntype"))
  do i = 1, nphotonband
    mapPatternType(i) =  mapPatternTypeString(i:i)
  enddo
  npar = parameterCount("ctilongtermpar")
  do i = 0, npar-1
    corrpar_in(i+1) = realParameter("ctilongtermpar",i)
  enddo

!  do i = 1,nphotonband
!    print*,i,mapPatternType(i),lothresh(i),hithresh(i)
!  enddo

  write(errstr,*) 'Test energy width of 1 eV       ',twidth
  call message(AppMsg,Noisy)
  write(errstr,*) 'Pixel randomisation on:         ',prandom
  call message(AppMsg,Noisy)
  write(errstr,*) 'Energy randomisation on:        ',erandom
  call message(AppMsg,Noisy)
  if (enable_oote) then
    write(errstr,*) 'Creation of out-of-time events: ',do_oote
    call message(AppMsg,Noisy)
  endif

end subroutine getInParamValues  


! ===========
! Declare DAL
! ===========
subroutine declareDal(ithp,prandom,erandom,intemperature,gainctiaccuracy, &
                      twidth, do_oote, quadrant, ccdId, withphotonmap)

  real(kind=double)  , intent(out)        :: intemperature
  integer(kind=Int32)                     :: gainctiaccuracy
  integer(kind=Int16)                     :: ithp
  logical, intent(in)                     :: prandom, erandom, twidth, &
                                             do_oote, withphotonmap
  character(len=8)                        :: extname
  integer(kind=Int16), intent(out)        :: quadrant, ccdid
  type(ColumnT)                           :: rawX, rawY, pha, pha_cti


  call message(AppMsg,Noisy,"... entering declareDal...")

!
! define infile
!
  infile = dataSet(stringParameter("eventset"),READ)

!
! check infile contains events extension
!
  write(extname,'(a)') "EVENTS"
  if (.not. hasBlock(infile,extname)) then    ! no EVENTS extension
    call error("noEvents","No EVENTS extension in input data set")
  endif

!
! check if infile contains EXPOSURE extension
!
  write(extname,'(a)') "EXPOSURE"
  if (.not. hasBlock(infile,extname)) then    ! no EXPOSURE extension
    call error("noExposure","No exposure extension in input data set")
  endif

  ! release input file

  call release(infile)

  ! define photonmapset
  
  eventmapset = stringParameter("photonmapset")
  if (withphotonmap) then
    call message(AppMsg,Noisy,"Photon maps will be created: "//trim(eventmapset))
  endif

  ! copy inset, open outset

  if (do_oote) then
!   outfilename = stringParameter("outoftimeset")
    outfilename = stringParameter("outset")
  else
    outfilename = stringParameter("outset")
  endif
  call copyDataSet(stringParameter("eventset"),outfilename)

  outfile = dataSet(outfilename,MODIFY)

  lburst = ( trim(stringAttribute(outfile,"DATAMODE")) == 'BURST' )
  if (lburst) then
    call message(AppMsg,Verbose,& 
                     "DATAMODE keyword value BURST will be changed to TIMING.")
    call setattribute(outfile,"DATAMODE","TIMING", &
                     "Instrument mode (changed from BURST to TIMING)")
  endif

  ! -----------------------------
  ! Add to events table extension
  ! -----------------------------

  events = table(outfile,"EVENTS",MODIFY)

  if (hasAttribute(events,"DATAMODE")) then
    linstmim = (trim(stringAttribute(events,"DATAMODE")) == "IMAGING")
  else
    call message(AppMsg,Sparse,'EVENTS extension dose not have keyword DATAMODE')
    linstmim = .false.
  endif
  if (linstmim) frameCTI = .true.

  if (hasAttribute(events,"F1128AVE")) then
    if (real64attribute(events,"F1128AVE") > -273.15) then
      intemperature = real64attribute(events,"F1128AVE")+273.15
    else
      intemperature = 183.2
    endif
  else
    intemperature = 183.2
  endif
  write(errstr,'(a,f8.3,a)') ' CCD temperature is: ',intemperature,' [K]'
  call message(AppMsg,Verbose)

  call setAttribute(events, "REEM_THR",ithp, & 
       comment="minimum PHA of preceding event for reemission check") 
  call setAttribute(events, "RANDOM_P",prandom, "randomize pixel position") 
  call setAttribute(events, "RANDOM_E",erandom, "randomize amplitudes in bin") 
  call setAttribute(events, "TWIDTH",twidth, "non-standard energy bin width")
  call setAttribute(events, "CCD_TEMP",intemperature,"K", &
       comment="CCD temperature")
  call setAttribute(events, "CTI_ACCU",gainctiaccuracy, &
       comment="accuracy of gain/cti correction")

  if (enable_SCR_88) then
    call setEventAttribute(events,EVATT_INVALID_PATTERN)
  endif
  call setEventAttribute(events,EVATT_CLOSE_TO_CCD_WINDOW)
  call setEventAttribute(events,EVATT_CLOSE_TO_ONBOARD_BADPIX)
  call setEventAttribute(events,EVATT_CLOSE_TO_BRIGHTPIX)
  call setEventAttribute(events,EVATT_CLOSE_TO_DEADPIX)
  call setEventAttribute(events,EVATT_COSMIC_RAY)
  call setEventAttribute(events,EVATT_TRAILING)
  call setEventAttribute(events,EVATT_ON_BADPIX)
  call setEventAttribute(events,EVATT_SECONDARY)

  write(errstr,*) 'Out_of_FOV: [3] IMAGING = ',linstmim,'  ENABLE = ', enable_out_of_fov
  call message(AppMsg,Noisy)
  if (linstmim) then
    if (enable_out_of_fov) then
      call setEventAttribute(events,EVATT_OUT_OF_FOV)
      write(errstr,*) 'Out_of_FOV: header keyword via setEventAttribute()'
      call message(AppMsg,Noisy)
    endif
  endif

  dims(1)=1
  detX     = addColumn(events,"DETX",Integer16,"0.05 arcsec",dims, &
             comment="Linearised Camera X-Coordinate of Event")
  call setAttribute(detX, "TLMIN",-18300, comment="Minimum for detX")
  call setAttribute(detX, "TLMAX",14000, comment="Maximum for detX")
  detY     = addColumn(events,"DETY",Integer16,"0.05 arcsec",dims, &
             comment="Linearised Camera Y-Coordinate of Event")
  call setAttribute(detY, "TLMIN",-17300, comment="Minimum for detY")
  call setAttribute(detY, "TLMAX",15400, comment="Maximum for detY")
  pha_cti  = addColumn(events,"PHA_CTI",Integer16,"channel",dims, &
             comment="CTI-corrected Event Energy")
  call setAttribute(pha_cti, "TLMIN",-32767, comment="Minimum for PHA_CTI")
  call setAttribute(pha_cti, "TLMAX",32767, comment="Maximum for PHA_CTI")
  call setNullValue(pha_cti,0)
  pi       = addColumn(events,"PI",Integer16,"eV",dims, &
             comment="Corrected Event Energy")
  call setAttribute(pi, "TLMIN",    0, comment="Minimum for PI")
  call setAttribute(pi, "TLMAX",32767, comment="Maximum for PI")
  call setNullValue(pi,-32768)
  pat_typ  = addColumn(events,"PAT_TYP",Integer16,"",dims, &
             comment="Pattern Type")  
  call setNullValue(pat_typ,32767)
  pat_ind  = addColumn(events,"PAT_IND",Integer16,"",dims, &
             comment="Pattern Index")
  call setNullValue(pat_ind,32767)
  pat_ori  = addColumn(events,"PAT_ORI",Integer16,"",dims, &
             comment="Pattern Orientation")  
  call setNullValue(pat_ori,32767)  
  pat_id   = addColumn(events,"PAT_ID",Integer16,"",dims, &
             comment="Combined pattern info 1")
  call setNullValue(pat_id,32767)
  pat_seq  = addColumn(events,"PAT_SEQ",Integer8,"",dims, &
             comment="Pattern Sequence")
  call setNullValue(pat_seq,255)
  pattern  = addColumn(events,"PATTERN",Integer8,"",dims, &
             comment="Combined pattern info 2")
  call setNullValue(pattern,13)

  ! -------------------------
  ! Open table columns needed
  ! -------------------------

  rawX = column(events,"RAWX",READ)
  rawY = column(events,"RAWY",READ)
  pha = column(events,"PHA",READ)
  qualflag = column(events,EVATT_COLUMN_NAME,MODIFY)  ! EVATT_COLUMN_NAME = FLAG

  ! set pointers
  rawXCol     => int16Data(rawX)
  rawYCol     => int16Data(rawY)
  phaCol      => int16Data(pha)
  qualflagCol => Int32Data(qualflag) 
  pha_ctiCol  => int16Data(pha_cti)
  piCol       => int16Data(pi)
  patternCol  => int8Data(pattern)
  pat_idCol   => int16Data(pat_id)
  pat_typCol  => int16Data(pat_typ)
  pat_indCol  => int16Data(pat_ind)
  pat_oriCol  => int16Data(pat_ori)
  pat_seqCol  => int8Data(pat_seq)

  ! - - - - - - - - - - - - - - - - - - - - - -
  ! Get overall CCD number from header keywords
  ! - - - - - - - - - - - - - - - - - - - - - - 

  quadrant = int16Attribute(events,"QUADRANT")
  ccdid    = int16Attribute(events,"CCDID")

  if (quadrant /= 0 .and. quadrant /= 1 .and.  &
      quadrant /= 2 .and. quadrant /= 3) then
!   call deleteDataSet(stringParameter("outset"))
    call deleteDataSet(outfilename)
    call error("invalidQuad","Invalid quadrant identifier")
  endif

  if (ccdid /= 0 .and. ccdid /= 1 .and. ccdid /= 2) then
    call deleteDataSet(outfilename)
    call error("invalidCCD","Invalid CCD identifier")
  endif

  ccd = quadrant*3+ccdid+1

  write(errstr,"(a,i1,a,i1,a,i2)") "working on CCD(",quadrant,"-",ccdid, &
                                ") = CCD No.",ccd
  call message(AppMsg,Verbose)

end subroutine declareDal


! ===========================
! Attribute parameter reading
! ===========================
subroutine getAttrParamValues(nxdim,nydim,nx0,ny0,srcpos,iamip,eamipsel, &
                              tstart,tstop)

  integer, intent(out)          :: nxdim,nydim,nx0,ny0,srcpos,iamip,eamipsel
  real(kind=double),intent(out) :: tstart, tstop
!

  character(len=20)      :: cmipmode
  character(len=12)      :: ECACHq_GAINCAME = "HIGH"

  call message(AppMsg,Noisy,"... entering getAttrParamValues...")

  ! ------------------------------------------
  ! Read Attributes from primary header of set
  ! ------------------------------------------

  if (hasAttribute(events,"GAINCAME")) then
    ECACHq_GAINCAME = trim(stringAttribute(events,"GAINCAME"))
    if (ECACHq_GAINCAME(1:4) .ne. "HIGH") then
      write(errstr,*) 'CCD',ccdNO,'is not in HIGH gain mode: ',&
                     trim(ECACHq_GAINCAME)
      call warning("CCDlowGain")
    endif
!   call message(AppMsg,Sparse,"Gain value = "//ECACHq_GAINCAME)
  endif

  instmode = int16Attribute(events,"CCDSMODE")
  lburst = ( trim(stringAttribute(events,"DATAMODE")) == modeID(BURST) )

  if (lburst) call setattribute(events,"DATAMODE",modeID(TIMING), &
                        "Instrument mode (changed from BURST to TIMING)")

  linstmim = .false.
  linstmti = .false.
  linstmbu = .false.

  select case (instmode)

    case(PRIME_FULL_WINDOW) ! IMAGING
      linstmim = .true. 

    case(PRIME_FULL_WINDOW_EXTENDED) ! IMAGING
      linstmim = .true. 

    case(PRIME_LARGE_WINDOW) ! IMAGING
      linstmim = .true. 

    case(PRIME_SMALL_WINDOW) ! IMAGING
      linstmim = .true. 

    case(FAST_TIMING) ! TIMING
      linstmti = .true. 

    case(FAST_BURST) ! BURST
      linstmbu = .true. 

    case default
      write(errstr,*) "InstrumentSubMode = ", instmode
      call message(AppMsg,Sparse)
      call deleteDataSet(outfilename)
      call error("badMode","Mode not supported")

  end select

  write(errstr,*) 'Out_of_FOV: [4] IMAGING = ',linstmim,'  ENABLE = ', enable_out_of_fov
  call message(AppMsg,Noisy)


  nevent = numberOfRows(events)
  write(errstr,'(1x,a,i11)') "Number of events in the data set:  ", nevent
  call message(AppMsg,Verbose)

  if (nevent == 0) call message(AppMsg,Sparse, &
                              "there are no events in the data set")

  ! --------------------------------
  ! Get the window size and position
  ! --------------------------------

  nxdim = int16Attribute(events,"WINDOWDX")
  nydim = int16Attribute(events,"WINDOWDY")
  nx0 =   int16Attribute(events,"WINDOWX0") 
  ny0 =   int16Attribute(events,"WINDOWY0")     

  write(errstr,*) "Window size: ",nxdim,"x",nydim," pixels"
  call message(AppMsg,Noisy)
  write(errstr,"(a,i2,a,i3,a)") &
    " Bottom left corner of readout window: (RAWX,RAWY) = (",nx0,",",ny0,")"
  call message(AppMsg,Noisy)

  ! -----------------------------------------
  ! Determine window edges and extended edges
  ! -----------------------------------------

  ixst  = nx0
  ixen  = ixst + nxdim - 1
  iyst  = ny0
  iyen  = iyst + nydim - 1
  ixste = ixst - 1
  ixene = ixen + 1
  iyste = iyst - 1
  iyene = iyen + 1

  ! --------------------------------------------------------------------
  ! Source position for TIMING and BURST mode (in pixels from the CAMEX)
  ! --------------------------------------------------------------------

  if (linstmti .or. linstmbu) & ! TIMING or BURST
                       srcpos =  int16Attribute(outfile,"SRCPOS")

! print*,'SRCPOS = ',srcpos

  ! --------------------------------------------------------
  ! Get MIP rejection methode and threshold used in epframes
  ! --------------------------------------------------------

  if (nevent > 0) then
    iamip    = int16Attribute(events,"MIP_THRS")
    cmipmode = stringAttribute(events,"MIP_METH")
    eamipsel = int16Attribute(events,"AN_EAMIP")
  else
    cmipmode = "NONE"
  endif

  lmipm =.false.
  if (cmipmode == "NONE") lmipm(1) = .true. ! mip columns not flagged
  if (cmipmode == "SAS")  lmipm(2) = .true. ! mip columns flagged
  if (cmipmode == "COM")  lmipm(3) = .true. ! mip columns and neighbours flagged
  if (cmipmode == "ONBOARD") lmipm(4) = .true.   ! similar to "COM"

  if (.not.lmipm(1) .and. .not.lmipm(2) .and. .not.lmipm(3) .and. &
                          .not.lmipm(4) ) then
    call deleteDataSet(outfilename)
    call error("badmipMode","Mip Mode not supported")
  endif

  if (nevent > 0) then
    tstart = real64Attribute(events,"TSTART")
    tstop  = real64Attribute(events,"TSTOP")
  else
    tstart = 0
    tstop  = 0
  endif

end subroutine getAttrParamValues



! ===============================
! Read and convolve bad pixel map
! ===============================
subroutine bpxmap

  integer                         :: i,j,ix,iy,ntbadrow

  character(len=8)                :: extname

  call message(AppMsg,Noisy,"... entering bpxmap...") 

  i2map = 0           ! Initialize badpix array
  i2mpc = 0           ! Initialize convolved badpix array
  ! -----------------------------------------------
  ! Does BADPIX extension exist in output data set?
  ! -----------------------------------------------

  write(extname,'(a)') "BADPIX"        

  if (.not. hasBlock(outfile,extname)) then    ! no BADPIX extension
    call warning("noBadpix","No badpix extension in output data set")

  else
    ! -------------------------
    ! Open the badpix extension
    ! -------------------------
    badpix = table(outfile,extname)

    badrawX = column(badpix,"RAWX",READ)
    badrawY = column(badpix,"RAWY",READ)
    yextent = column(badpix,"YEXTENT",READ)
    badPixType = column(badpix,"TYPE",READ)
    badflag = column(badpix,"BADFLAG",READ)

    badrawXCol => int16Data(badrawX)
    badrawYCol => int16Data(badrawY)
    yextentCol => int16Data(yextent)
    badPixTypeCol => int16Data(badPixType)
    badflagCol => int16Data(badflag)

    ntbadrow = numberOfRows(badpix)

    do i = 0, ntbadrow - 1
      ix = badrawXCol(i)
      do j = 1, yextentCol(i)
        iy = badrawYCol(i) + j - 1

   ! check if pixel inside window
        if (iy >= iyst .and. iy <= iyen) then
   ! Type = 1 -> Hot/bright    BadFlag = 1 -> Uploaded    = 3 -> badpixfind
   ! Code: TYPE * 100 + BADFLAG * 10
          i2map(ix,iy) = badPixTypeCol(i) * 100 + badflagCol(i) * 10
        else
          write(errstr,*) "Bad Pixel outside CCD window in y: ",iy
          if (iy > 1) then
!           call warning("wrongBadpixel")
            if (verbosity(AppMsg) > Verbose) call message(AppMsg,Verbose)
          else
            if (verbosity(AppMsg) > Noisy) call message(AppMsg,Noisy)
          endif
        endif

      enddo
    enddo

    ! release BADPIX extension to save memory

    call release(badpix)

  endif

  ! --------------------------  
  ! Convolve the bad pixel map
  ! --------------------------
  ! i.e. mark the pixels above, below, left and right of a bad
  ! pixel; if a single or the maximum of a pattern coincides with such 
  ! a pixel, the single or pattern will be considered as invalid; 
  ! edge pixels will be treated as if there was a bad pixel just outside 
  ! the sensitive area of the CCD


  ! area around bad pixels:

  do iy = iyst, iyen
    do ix = ixst, ixen
      if (i2map(ix,iy) >= 1) then
        ! Code: TYPE * 10 + BADFLAG  
        i2mpc(ix,iy)      =  i2map(ix,iy)
        i2mpc(ix-1,iy)    = &
          max(i2mpc(ix-1,iy),i2map(ix,iy)/10)
        i2mpc(ix,iy-1)    = &
          max(i2mpc(ix,iy-1),i2map(ix,iy)/10)
        i2mpc(ix+1,iy)= &
          max(i2mpc(ix+1,iy),i2map(ix,iy)/10)
        i2mpc(ix,iy+1)= &
          max(i2mpc(ix,iy+1),i2map(ix,iy)/10)
      endif
    enddo
  enddo

  ! edge pixels:

  i2mpc(ixst,iyst:iyen) = max(i2mpc(ixst,iyst:iyen),1)
  i2mpc(ixen,iyst:iyen) = max(i2mpc(ixen,iyst:iyen),1)

  i2mpc(ixst:ixen,iyst) = max(i2mpc(ixst:ixen,iyst),1)
  i2mpc(ixst:ixen,iyen) = max(i2mpc(ixst:ixen,iyen),1)

  ! copy i2map and i2mpc in frame specific maps
  i2mapf = i2map
  i2mpcf = i2mpc
  ! initialize li2mapold and li2mpcold
  li2mapold = .false.
  li2mpcold = .false.

end subroutine bpxmap



! ==============================================================
! Read Frameinfo (number of frames, boundaries, discarded lines)
! Real relevant attributes from exposure extension 
! ==============================================================
subroutine readFrames(temperature_gain)

  integer(kind=Int32), dimension(:),pointer :: framelimCol
  integer(kind=Int32)                       :: expos_evtnumber,i,f,frmold
  real(kind=single),intent(out)             :: temperature_gain

  call message(AppMsg,Noisy,"... entering readFrames...") 

  ! ---------------------------
  ! Open the EXPOSURE extension
  ! ---------------------------
  exposure = table(outfile,"EXPOSURE")
  nfrm = numberOfRows(exposure)
  write(errstr,"(1x,a,i11)") "Total number of frames [FRAMETOT]: ",&
         int32Attribute(events,"FRAMETOT")
  call message(AppMsg,Verbose)
  write(errstr,"(1x,a,i11)") "Total number of frames [EXPOSURE]: ",nfrm 
  call message(AppMsg,Verbose)

  framelim = column(exposure,"FRAMELIM",READ)
  framelimCol => int32Data(framelim)
!
! check EXPOSURE extension: should be monotonuous in FRAMELIM
!
  frmold = 0
  do i = 1,nfrm-1
    f = framelimCol(i)
    if (f < frmold) then
      write(errstr,*) "Extra frames at end after number: ",i, &
                      ",   frames to be ignored: ",nfrm-i
      call message(AppMsg,Sparse)
      nfrm = i
      exit
    endif
    frmold = f
  enddo

!
! FRAMELIM contains the number of the last event recorded
! before the end of the frame
!
  allocate(ind(0:nfrm+1))
  ind(0) = 0
  ind(1:nfrm) = framelimCol(0:nfrm-1)
  expos_evtnumber = ind(nfrm)

  ndsclin1 = column(exposure,"NDSCLIN1",READ)
  ndsclin2 = column(exposure,"NDSCLIN2",READ)
  ndsclin3 = column(exposure,"NDSCLIN3",READ)
  ndsclin4 = column(exposure,"NDSCLIN4",READ)

  ndsclin1Col => int16Data(ndsclin1)
  ndsclin2Col => int16Data(ndsclin2)
  ndsclin3Col => int16Data(ndsclin3)
  ndsclin4Col => int16Data(ndsclin4)

  allocate(ndsclin(0:3,0:nfrm))

  ndsclin(0,0:nfrm-1) = ndsclin1Col(0:nfrm-1)
  ndsclin(1,0:nfrm-1) = ndsclin2Col(0:nfrm-1)
  ndsclin(2,0:nfrm-1) = ndsclin3Col(0:nfrm-1)
  ndsclin(3,0:nfrm-1) = ndsclin4Col(0:nfrm-1)

  if (expos_evtnumber == nevent-1) then
    write(errstr,'(a,a)') &
         'Last event frame seems to be missing in EXPOSURE extension. ', &
         'This does not influence pattern and amplitude information. '
    call message(AppMsg,Sparse)
    nfrm = nfrm + 1
    ind(nfrm) = nevent 
    ndsclin(0:3,nfrm-1) = 0
  endif
!
! read relevant temperature to correct for gain-temperature variation
!
  if (hasAttribute(exposure,"TQB1_AVE")) then
    temperature_gain = real32Attribute(exposure,"TQB1_AVE")
  else
    temperature_gain = -9999.0
  endif

!
! release EXPOSURE extension to save memory
!
  call release (exposure)

end subroutine readFrames



! ===========================
! Correct for discarded lines
! ===========================
subroutine discLines(ifrm)

  implicit none

  integer                         ifrm, i, ii, i0, i1, ic, n, k
  logical                         lndscli(0:65), ldisc

  ! copy bad pixel map to frame specific map
  ! copy modified bad pixel map to frame specific map
  ! check for mip rejection mode and correct accordingly
  ! flag mip columns

  ! first initialize i2mapf and i2mpcf by resetting columns
  ! changed in last run by columns from original maps

  if (verbosity(AppMsg) >= Noisy) then

    call message(AppMsg,Noisy,"... entering discLines...") 

    write(errstr,*) " li2mapold"," .",li2mapold,"."   ! 0,1-64,65
    call message(AppMsg,Noisy)
    write(errstr,*) " li2mpcold",li2mpcold            ! 0-65
    call message(AppMsg,Noisy)

  endif

  do i = 1, 64
    if (li2mapold(i)) then 
      i2mapf(i,:) = i2map(i,:)
      li2mapold(i) = .false.
    endif
    if (li2mpcold(i)) then
      i2mpcf(i,:) = i2mpc(i,:)
      li2mpcold(i) = .false.
    endif
  enddo

  lndscli = .false.
  ldisc   = .false.

  do i = 0, 3
    if (ndsclin(i,ifrm) /= 0) then
      ldisc = .true.
      n = i * 16
      do k = 1, 16
        ii = k + n
        lndscli(ii) = (ibits(ndsclin(i,ifrm),k-1,1) > 0)
      enddo
    endif
  enddo

  if (verbosity(AppMsg) >= Noisy) then
    write(errstr,*) " lndscli  ",lndscli
    call message(AppMsg,Noisy)
  endif

  if (ldisc) then
    ic = -2
    do i = 1, 64
      if (lndscli(i)) then
        i2mapf(i,:)  = i2map(i,:) + 10000
        li2mapold(i) = .true.
        i2mpcf(i,:)  = i2mpc(i,:) + 10000
        li2mpcold(i) = .true.
        if (lmipm(2)) then
          ! flag neighbouring columns like edge pixels
          i0 = i - 1
          i1 = i + 1
          if (i0 /= ic) then
            if (lndscli(i0)) &
              i2mpcf(i0,:)  = max(i2mpc(i0,:),1)
              li2mpcold(i0) = .true.
          endif
          ic = i1
          if (lndscli(i1)) &
            i2mpcf(i1,:)  = max(i2mpc(i1,:),1)
            li2mpcold(i1) = .true.
        endif
      endif
    enddo
  endif

  if (verbosity(AppMsg) >= Noisy) then
    write(errstr,*) " li2mapnew"," .",li2mapold,"."
    call message(AppMsg,Noisy)
    write(errstr,*) " li2mpcnew",li2mpcold
    call message(AppMsg,Noisy)
  endif

end subroutine discLines



! =======================
! Trailing event analysis
! =======================
subroutine trailEvent(nevf,ixev,iyev,ampl,iqual,lqual,thp,itrail)

  integer,             intent(in)   :: nevf

  integer(kind=Int16),dimension(:)  :: ixev, iyev

  real(kind=single),dimension(:)    :: ampl

  integer(kind=Int32),dimension(:)  :: iqual

  logical,dimension(:)              :: lqual

  real(kind=single),   intent(in)   :: thp

  integer(kind=Int16),intent(inout) :: itrail

  real(kind=double)                 :: amp,h

  integer(kind=Int32)               :: i, k

  logical                           :: lstop, lpre, lsl

  if (verbosity(AppMsg) >= Noisy) &
      call message(AppMsg,Noisy,"... entering trailEvent...") 

  itrail = 0

  do i = 1, nevf
    if (lqual(i)) then
      k  = i
      lstop = .false.

      ! determine the distance to the preceding event:
      do while (.not.lstop .and. k > 1)
        k = k - 1
        if(lqual(k)) lstop = (ixev(k) == ixev(i) .and. iyev(k) < iyev(i))
      enddo

      ! check if immediately preceding
      if (lstop) then
        lpre = (iyev(k) == iyev(i) - 1)
      else
        lpre = .false.
      endif

      ! ampl(i)  == amplitude of trailing event
      ! ampl(k)  == amplitude of main (preceding) event

      lsl = .true.

      if (lpre) then
        if (verbosity(AppMsg) >= Noisy) then
          write(errstr,*)"Main event (nevf,x,y,ampl):      ", &
                k,ixev(k),iyev(k),ampl(k)
          call message(AppMsg,Noisy)
          write(errstr,*)"Trailing event (nevf,x,y,ampl):  ", &
                i,ixev(i),iyev(i),ampl(i)
          call message(AppMsg,Noisy)
        endif
        ! check if preceding event is above threshold for reemission check
        if (ampl(k) >= thp) then      ! threshold for preceding events

          ! calculate maximal expected reemission amplitude
          amp = ampl(k)
!!! CAL
          h = CAL_pnReEmissionThreshold(amp, iyev(i))
!!! KOD
!!!       h = 1.25E+01 - 4.5E-04 * amp + (5.4E-02 + 3.5E-05 * amp)*(iyev(i)-1)
!!!       print*,CAL_pnReEmissionThreshold(amp, iyev(i)),h
          if (verbosity(AppMsg) > Noisy) then
            write(errstr,*)"Reemission threshold:      ", h
            call message(AppMsg,Noisy)
          endif
          lsl = (ampl(i) >= h)               ! threshold for trailing events
        endif
      endif

      ! flag and count detected reemission events
      if (.not.lsl) then
        iqual(i) = ior(iqual(i), EVATT_TRAILING)
        lqual(i) = .false.
        itrail = itrail + 1                 ! counting rejected events
        if (verbosity(AppMsg) >= Noisy) then
          write(errstr,*)"Trailing event (nevf, ampl):  ", i , ampl(i)
          call message(AppMsg,Noisy)
        endif
      endif

    endif
  enddo
  if (verbosity(AppMsg) >= Noisy) then
    write(errstr,*)"Number of trailing event in frame:  ", itrail
    call message(AppMsg,Noisy)
  endif

end subroutine trailEvent



! ================
! Pattern analysis
! ================
subroutine split(nevf,ixev,iyev,ampl,iqual, &
                   lqual,amip,ipat1,ipat2,ipat3,ipat4,i4dist,n1,n2,n3,n4, &
                   nrest,nmax,nmip,do_oote)

  implicit none

  integer,            intent(in)       :: nevf
  real(kind=single),  intent(in)       :: amip
  logical,            intent(in)       :: do_oote

  integer(kind=Int16),dimension(:),intent(in)    :: ixev,iyev
  real(kind=single)  ,dimension(:),intent(in)    :: ampl
  logical            ,dimension(:),intent(in)    :: lqual
  integer(kind=Int32),dimension(:),intent(inout) :: iqual  ! ,jqual
  integer(kind=Int16),dimension(:),intent(out)   :: ipat1,ipat2,ipat3,ipat4
  integer,                         intent(inout) :: i4dist(3,3), &
                                                    n1,n2,n3,n4,nrest,nmax,nmip

  integer                :: k, k1, k2, k3, kx, ky, &
                            i, j, imin, ix, iy, &
                            ipat, ngev, keqv, neqv, kpat, mpat, &
                            ix0, iy0, kmax, ixs, iys, &
                            neqvl, keqvl,maxeqv_old,&
                            nsel, isel, &
                            i4h(3,3), itype, code, sevf = 0

  integer,dimension(:,:),allocatable    :: ieqv,ieqvl
  integer,dimension(:),allocatable      :: ind1,ind2,ixx,iyy,npnt, &
                                           ncls,inpat,ixsel, iysel,izsel
  integer(kind=int16),dimension(:),allocatable :: jpat1,jpat2,jpat3,jpat4
  logical,dimension(:),allocatable      :: lind,lrec,lval

  real(kind=single)      :: r4buf(3,3), z, zmax

  logical                :: l1, l2, l3, l4, &
                            lfound, lmip, lmin, lflag, lwarn, linvalid
  logical, parameter     :: lscr = .false.

  character(len=1)       :: ct(1:EPN_MAX_X,1:EPN_MAX_Y)
  character(len=80)      :: chbuf

  if (verbosity(AppMsg) >= Noisy) &
      call message(AppMsg,Noisy,"... entering split...") 

  ! allocate array dimensions
  allocate(ieqv(1:maxeqv,1:2))
  allocate(ieqvl(1:maxeqv,1:maxcls))
  allocate(ind1(1:maxpat))
  allocate(ind2(1:maxpat))
  allocate(ixx(1:nevf))
  allocate(iyy(1:nevf))
  allocate(npnt(1:maxpat))
  allocate(ncls(1:maxcls))
  allocate(inpat(1:maxsel))
  allocate(ixsel(1:maxsel))
  allocate(iysel(1:maxsel))
  allocate(izsel(1:maxsel))
  allocate(lind(1:maxpat))
  allocate(lrec(1:maxpat))
  allocate(lval(1:maxpat))
  if (lscr) then
    allocate(jpat1(1:nevf),jpat2(1:nevf),jpat3(1:nevf),jpat4(1:nevf))
    jpat1(1:nevf) = 0
    jpat2(1:nevf) = 0
    jpat3(1:nevf) = 0
    jpat4(1:nevf) = 0
  endif

! print*,'nevf,maxeqv,maxcls,maxpat,maxsel = ',nevf,maxeqv,maxcls,maxpat,maxsel
! print*,ixev(1:nevf),iyev(1:nevf),ampl(1:nevf),iqual(1:nevf),lqual(1:nevf),amip
!
! initialize the maps with the amplitudes
!
! fill the maps with the amplitudes and pointers and check for bad pixels
!

  ngev = 0           ! counter for good events in frame
  sevf = sevf+nevf   ! counter for total events to be processed at end
  maxeqv_old = nevf / 2 + 1
  lwarn = .true.
  j = 0

  do k = 1, nevf 
!   print*,'k, LQUAL(k) =',  k, LQUAL(k) 
!
! initialize pattern arrays (ipat4 is set for all events in the program)
!
    ipat1(k) = 0    ! array for the pattern type    
    ipat2(k) = 0    ! array for the pattern index
    ipat3(k) = 0    ! array for the pattern orientation
!
! if quality is better than certain limit (LQUAL==.true.) 
!
    if (.not.lqual(k)) then
      cycle  
    else        
      ix = ixev(k) 
      iy = iyev(k)
!     print*,'IX,IY,I2MAPF(IX,IY) = ',IX,IY,I2MAPF(IX,IY)
      if (i2mapf(ix,iy) == 0) then
        i4map(ix,iy) = -1         ! 0: no event, -1: event
        indmap(ix,iy) = k
        ngev = ngev + 1
      else
!
! check for events in bad pixels and set flag:
! give warning, if found bad pixel is uploaded.
! This should be changed to error, as soon as correct
! time handling is implemented in CAL and CCF
!
        if (i2mpcf(ix,iy) > 100) then
          iqual(k) = ior(iqual(k),EVATT_ON_BADPIX)
          if (mod(i2mpcf(ix,iy),100) == 10) then       
            write(errstr,"(a,i9,a,3i4)") &
                 "Event is on an uploaded bad pixel: frame number:", &
                  ifrm, "  CCD pixel address:", ccdNo,ix,iy
            if (do_oote) then
              call message(AppMsg,Verbose)
            else
              call warning("badEvent")
            endif
          endif
        endif

      endif
    endif
  enddo

  if (verbosity(AppMsg) >= Noisy) then
    write(errstr,"(a,2i8)") "SPLIT: frame number, good events in frame:", &
                            ifrm,ngev
    call message(AppMsg,Noisy)
  endif

!
! determine the distances to neighbouring events along each column
! (towards the CAMEX; 0 means: first pixel to be read in a column)
! and fill the ipat4 array for the pattern sequence:
! determine if member of pattern and first pattern no:
!
  lfound = .false.                ! if true at least one pattern
  nsel = 0
  if (ngev >= 2) then             ! only if more than one event in frame
    do i= nevf, 1, -1
      ix0 = ixev(i)
      iy0 = iyev(i)
      if (i4map(ix0,iy0) /= 0) then
        lflag = .false.
        l1 = (i4map(ix0-1,iy0) /= 0)  ! neighbour is left
        l2 = (i4map(ix0+1,iy0) /= 0)  ! neighbour is right
        l3 = (i4map(ix0,iy0-1) /= 0)  ! neighbour is below
        l4 = (i4map(ix0,iy0+1) /= 0)  ! neighbour is above
        if (l1.or.l2.or.l3.or.l4) then
          i4map(ix0,iy0) = -2         ! doubles or greater pattern (singles have -1)
          lfound = .true.

          ! store pointers to pattern pixels in reverse order into
          !   inpat, ixsel, iysel and determine number of pattern pixels

          nsel = nsel + 1             ! nsel = number of pixels with neighbours
          if (nsel > maxsel) then
            ! maximum allowed number of pattern pixels is exceeded
            ! increase parameter maxsel 
            write(errstr,"(a,i8)") &
                  "Maximum number of pattern pixels exceeded:", maxsel
            call message(AppMsg,Sparse)
            call message(AppMsg,Sparse,"increase parameter maxsel")
            call deleteDataSet(outfilename)
            call error("tooManySelected", &
                  "Number of selected pattern pixels exceeds the limit")     
          endif

          inpat(nsel) = i             ! initial pattern number
          ixsel(nsel) = ix0
          iysel(nsel) = iy0
        endif
        k = i - 1
        if (k >= 1) then
          ix = ixev(k)
          iy = iyev(k)      
          do while (k > 1)
            if (ix == ix0 .and. i4map(ix,iy) /= 0) then
              ipat4(indmap(ix0,iy0)) = iy0 - iy    ! distance to next event in Y
              lflag = .true.
              k = 1
            else
              k= k - 1
              ix = ixev(k)
              iy = iyev(k)
            endif 
          enddo
          if (.not.lflag) then
            if (ix == ix0 .and. i4map(ix,iy) /= 0) then
              ipat4(indmap(ix0,iy0)) = iy0 - iy 
            else
              ipat4(indmap(ix0,iy0)) = 0
            endif
          endif
        else
          ipat4(indmap(ix0,iy0)) = 0
        endif
      endif
    enddo
  else
    do i = 1, nevf
      if (lqual(i)) ipat4(i) = 0
    enddo
  endif

  !
  ! initialize the statistical variables:
  !

  n1 = 0
  n2 = 0
  n3 = 0
  n4 = 0
  nrest = 0
  nmax = 0
  nmip = 0

  !
  ! initialize the distribution map:
  !

  i4dist = 0


  !
  ! DETERMINE THE PATTERNS:
  !

  if (lfound) then
    ! 
    !  write a preliminary pattern sequence into i4map (this sequence
    !  is preliminary because patterns with complex morphology may
    !  not be recognized as a single coherent pattern and may receive
    !  several pattern sequence numbers):
    !
    ipat = 1
    ixs = ixev(inpat(nsel))
    iys = iyev(inpat(nsel))
    i4map(ixs,iys) = ipat
    izsel(nsel) = ipat
    lwarn = .true.
    do k = nsel - 1, 1, -1
      ix = ixev(inpat(k))
      iy = iyev(inpat(k))
      k1 = i4map(ix-1,iy)
      k2 = i4map(ix,iy-1)
      k3 = 0
      kx = ix
      do while (i4map(kx+1,iy) == -2 .and. kx < ixen-1)
        kx = kx + 1
        if (i4map(kx,iy-1) > 0) cycle 
      enddo
      k3 = i4map(kx,iy-1)
      l1 = (k1 > 0)
      l2 = (k2 > 0)
      l3 = (k3 > 0)
      if (l1.or.l2.or.l3) then
        i4map(ix,iy) = max(k1,k2,k3)
      else
        ipat = ipat + 1
        i4map(ix,iy) = ipat
      endif
      izsel(k) = i4map(ix,iy)
    enddo

    npat = ipat
    ! check if maximum allowed pattern number is exceeded
    ! if yes set parameter maxpat to at least npat

    if (npat > maxpat) then
      call message(AppMsg,Sparse,'--------------------------------------------')
      write(errstr,"(a,i8)") "number of patterns found:          ", npat
      call message(AppMsg,Sparse)
!     write(errstr,"(a,i8)") "Maximum number of Patterns allowed:", maxpat
      write(errstr,"(a,i8)") "number of events in this frame:    ", maxpat
      call message(AppMsg,Sparse)
      call deleteDataSet(outfilename)
      
      call message(AppMsg,Sparse,'Event coordinates in this frame:')
      do k = 1,nevf
        write(errstr,"(a,4i8)") 'I,IX,IY,IA = ',k,ixev(k),iyev(k),nint(ampl(k))
        call message(AppMsg,Sparse)
      enddo
      
      call error("tooManyPattern","Number of patterns exceeds the limit")     
    endif


    !
    ! combine patterns which have received different sequence numbers:
    ! only if more than one pattern    
    !
    if (npat > 1) then 

      ! first step in y

      keqv = 0
      lwarn = .true.
      do k = nsel, 1, -1
        ix = ixev(inpat(k))
        iy = iyev(inpat(k))
        k1 = i4map(ix,iy)
        k2 = i4map(ix,iy-1)
        if (k2 > 0 .and. k2 /= k1) then
          if (keqv < maxeqv) then
            keqv = keqv + 1
            ieqv(keqv,1) = max(k1,k2)
            ieqv(keqv,2) = min(k1,k2)
            if (lwarn .and. keqv > maxeqv_old) then
              lwarn = .false.
              call message(AppMsg,Sparse,' ')
              write(errstr,"(a,i8)") &
                "Maximum number of equivalence classes in Y exceeded:",maxeqv_old
              call warning('veryManyClasses')

              ct(1:EPN_MAX_X,1:EPN_MAX_Y) = ' '
              call message(AppMsg,Sparse,' ')
              call message(AppMsg,Sparse,'Event coordinates in this frame:')
              call message(AppMsg,Sparse,'K1 = i4map(ix,iy), K2 = i4map(ix,iy-1):')

              do i = nsel, 1, -1
                ix = ixev(inpat(i))
                iy = iyev(inpat(i))
                ct(ix,iy) = '*'
                k1 = i4map(ix,iy)
                k2 = i4map(ix,iy-1)
                write(errstr,"(a,7i8)") 'IFRM,I,IX,IY,IA,K1,K2 = ',&
                        ifrm,i,ix,iy,nint(ampl(inpat(i))),k1,k2
                call message(AppMsg,Sparse)
              enddo
              write(errstr,*) 'Events of this frame end at index ',sevf
              call message(AppMsg,Sparse)
              call print_CCD_ascii(ct)
              lwarn = .false.
            endif
          else
            write(errstr,"(a,i8)") &
              "Maximum number of equivalence classes in Y exceeded:", maxeqv
            call deleteDataSet(outfilename)
            call error("tooManyClasses")     
          endif
        endif
      enddo
      neqv = keqv
!     write(errstr,*)' Number of event classes in this frame: ',neqv
!     call message(AppMsg,Sparse)

      !
      ! check if patterns have been found to combine
      !
      if (neqv > 0) then

        !
        ! produce equivalence classes:
        !
        call spleqv(neqv,ieqv,neqvl,ieqvl,ncls)

        !
        ! convert them into pointers:
        !

        k = 0
        do keqvl = 1, neqvl
          if (ncls(keqvl) > 0) then
            do i = 2, ncls(keqvl)
              k = k + 1
              ieqv(k,1) = ieqvl(keqvl,i)
              ieqv(k,2) = ieqvl(keqvl,1)
            enddo
          endif
        enddo
        neqv = k

        !
        ! reassign pattern numbers:
        !

        k = 0
        do ipat = 1, npat
          lfound = .false.
          keqv = 0
          do while (.not.lfound .and. keqv < neqv)
            keqv = keqv + 1
            lfound = (ipat == ieqv(keqv,1))
          enddo
          if (lfound) then
            ind1(ipat) = ieqv(keqv,2)
          else
            ind1(ipat) = ipat
          endif
        enddo

        lind(1:npat) = .false.

        do ipat = 1, npat
          lind(ind1(ipat)) = .true.
        enddo

        k = 0
        do ipat = 1, npat
          if (lind(ipat)) then
            k = k + 1
            ind2(ipat) = k
          else
            ind2(ipat) = 0
          endif
        enddo
 
        do ipat = 1, npat
          ind1(ipat) = ind2(ind1(ipat))
        enddo      

        !
        ! assign new pattern sequence numbers and define number of patterns:
        !

        mpat = 0
        do i = nsel, 1, -1
          ix = ixev(inpat(i))
          iy = iyev(inpat(i))
          k = i4map(ix,iy)
          do keqv = 1, neqv
            k1 = ieqv(keqv,1)
            k2 = ieqv(keqv,2)
            if (k == k1) then
              i4map(ix,iy) = ind1(k2)
            else 
              i4map(ix,iy) = ind1(k)
            endif
          enddo
          izsel(i) = i4map(ix,iy)
          mpat = max(mpat,i4map(ix,iy))
        enddo

        npat = mpat   
      endif

      ! now step in x

      lwarn = .true.
      keqv = 0
      do k = nsel, 1, -1
        ix = ixev(inpat(k))
        iy = iyev(inpat(k))
        k1 = i4map(ix,iy)
        k2 = i4map(ix+1,iy)
        if (k2 > 0 .and. k2 /= k1) then
          if (keqv < maxeqv) then
            keqv = keqv + 1
            ieqv(keqv,1) = max(k1,k2)
            ieqv(keqv,2) = min(k1,k2)
            if (lwarn .and. keqv > maxeqv_old) then
              lwarn = .false.
              call message(AppMsg,Sparse,' ')
              write(errstr,"(a,i8)") &
               "Maximum number of equivalence classes in Y exceeded:",maxeqv_old
              call warning('veryManyClasses')

              ct(1:EPN_MAX_X,1:EPN_MAX_Y) = ' '
              call message(AppMsg,Sparse,' ')
              call message(AppMsg,Sparse,'Event coordinates in this frame:')
              call message(AppMsg,Sparse,'K1 = i4map(ix,iy), K2 = i4map(ix+1,iy):')

              do i = nsel, 1, -1
                ix = ixev(inpat(i))
                iy = iyev(inpat(i))
                ct(ix,iy) = '*'
                k1 = i4map(ix,iy)
                k2 = i4map(ix+1,iy)
                write(errstr,"(a,7i8)") 'IFRM,I,IX,IY,IA,K1,K2 = ',&
                     ifrm,i,ix,iy,nint(ampl(inpat(i))),k1,k2
                call message(AppMsg,Sparse)
              enddo
              write(errstr,*) 'Events of this frame end at index ',sevf
              call message(AppMsg,Sparse)
              call print_CCD_ascii(ct)
              lwarn = .false.
            endif
          else
            write(errstr,"(a,i8)") &
              "Maximum number of equivalence classes in X exceeded:", maxeqv
            call deleteDataSet(outfilename)
            call error("tooManyClasses")     
          endif
        endif
      enddo
      neqv = keqv
     
      !
      ! check if patterns have been found to combine
      !
      if (neqv > 0) then

        !
        ! produce equivalence classes:
        !
        call spleqv(neqv,ieqv,neqvl,ieqvl,ncls)

        !
        ! convert them into pointers:
        !

        k = 0
        do keqvl = 1, neqvl
          if (ncls(keqvl) > 0) then
            do i = 2, ncls(keqvl)
              k = k + 1
              ieqv(k,1) = ieqvl(keqvl,i)
              ieqv(k,2) = ieqvl(keqvl,1)
            enddo
          endif
        enddo
        neqv = k

        !
        ! reassign pattern numbers:
        !

        k = 0
        do ipat = 1, npat
          lfound = .false.
          keqv = 0
          do while (.not.lfound .and. keqv < neqv)
            keqv = keqv + 1
            lfound = (ipat == ieqv(keqv,1))
          enddo
          if (lfound) then
            ind1(ipat) = ieqv(keqv,2)
          else
            ind1(ipat) = ipat
          endif
        enddo
        
        lind(1:npat) = .false.
       
        do ipat = 1, npat
          lind(ind1(ipat)) = .true.
        enddo

        k = 0
        do ipat = 1, npat
          if (lind(ipat)) then
            k = k + 1
            ind2(ipat) = k
          else
            ind2(ipat) = 0
          endif
        enddo
 
        do ipat = 1, npat
          ind1(ipat) = ind2(ind1(ipat))
        enddo      

        !
        ! assign new pattern sequence numbers and define number of patterns:
        !

        mpat = 0
        do i = nsel, 1, -1
          ix = ixev(inpat(i))
          iy = iyev(inpat(i))
          k = i4map(ix,iy)
          do keqv = 1, neqv
            k1 = ieqv(keqv,1)
            k2 = ieqv(keqv,2)
            if (k == k1) then
              i4map(ix,iy) = ind1(k2)
            else 
              i4map(ix,iy) = ind1(k)
            endif
          enddo
          izsel(i) = i4map(ix,iy)
          mpat = max(mpat,i4map(ix,iy))
        enddo
        npat = mpat
      endif
    endif
    ! end of pattern combination loop
  
!
! store patterns:
!

    k = 0
    do ipat = 1, npat
      ind1(ipat) = k + 1
      do isel = nsel, 1, -1
        if (izsel(isel) == ipat) then
          k = k + 1
          ixx(k) = ixsel(isel)
          iyy(k) = iysel(isel)
        endif
      enddo
      npnt(ipat) = k - ind1(ipat) + 1  
    enddo

!
! determine the distribution histogram:
!

    do ipat = 1, npat

      r4buf(1:3,1:3) = 0.0    ! accumulating array

      i4h(1:3,1:3) = 0    ! temporary array

      kmax = ind1(ipat)
      zmax = ampl(indmap(ixx(kmax),iyy(kmax)))
      do k = ind1(ipat) + 1, ind1(ipat) + npnt(ipat) - 1
        z = ampl(indmap(ixx(k),iyy(k)))
        if (z > zmax) then
          zmax = z
          kmax = k
        endif
      enddo
      ix0 = ixx(kmax)
      iy0 = iyy(kmax)
      z = 0.0
      do k = ind1(ipat), ind1(ipat) + npnt(ipat) - 1
        ix = ixx(k) - ix0
        iy = iyy(k) - iy0
        if (max(iabs(ix),iabs(iy)) <= 1) then
          i4h(ix+2,iy+2) = nint(ampl(indmap(ixx(k),iyy(k))))
        endif
      enddo

!
! check whether the pattern itself is recognizable:
! (ignoring bad and edge pixels)
!

      k = npnt(ipat)
      if (k <= 2) then       ! double
        lrec(ipat) = .true.
      else if (k == 3) then  ! triple
        lrec(ipat) = .true.
        if (max(i4h(1,1),i4h(1,3),i4h(3,1),i4h(3,3)) /= 0) then
          lrec(ipat) = .false.
        else
          if (i4h(1,2) /= 0 .and. i4h(3,2) /= 0) then
            lrec(ipat) = .false.
          else if (i4h(2,1) /= 0 .and. i4h(2,3) /= 0) then
            lrec(ipat) = .false.
          else
            i = 0
            do iy = 1, 3
              do ix = 1, 3
                if (i4h(ix,iy) > 0) i = i + 1
              enddo
            enddo
            if (i /= 3) lrec(ipat) = .false.
          endif
        endif
      else if (k == 4) then  ! quadruple
        lrec(ipat) = .true.
        imin = i4h(2,2)
        lmin = .false.
        do iy = 1, 3
          do ix = 1, 3
            i = i4h(ix,iy)
            if (i < imin .and. i > 0) then
              lmin = .true.                 ! changed  26.08.99 mjf
              imin = i
              kx = ix
              ky = iy
            endif
          enddo
        enddo

        if (.not.lmin) then                 ! changed  26.08.99 mjf
          lrec(ipat) = .false.              ! changed  26.08.99 mjf
        else if (.not.(i4h(1,1) == imin .or. i4h(1,3) == imin .or.   &
               i4h(3,1) == imin .or. i4h(3,3) == imin)) then
          lrec(ipat) = .false.
        else if (i4h(2,ky) == 0 .or. i4h(kx,2) == 0) then
          lrec(ipat) = .false.
        else
          i = 0
          do iy = 1, 3
            do ix = 1, 3
              if (i4h(ix,iy) > 0) i = i + 1
            enddo
          enddo
          if (i /= 4) lrec(ipat) = .false.
        endif
      else
        lrec(ipat) = .false.
      endif
!
! check patterns for the presence of bad or edge pixels
! and mark them as invalid, if the maximum amplitude of
! the pattern coincides with a pixel flagged in i2mpcf:
! (singles will be analysed separately)
!
      lval(ipat) = (i2mpcf(ix0,iy0) == 0) 
     
      code = 0
      if (.not.lval(ipat)) then
        code = i2mpcf(ix0,iy0)
      endif

!
! check for invalid patterns (SCR_88):
!
      linvalid = .false.
!
! check for MIPs:
!
      lmip = .false.
      kpat = ind1(ipat) - 1
      do while (kpat < ind1(ipat)+npnt(ipat)-1 .and. .not.lmip)
        kpat = kpat + 1
        if (ampl(indmap(ixx(kpat),iyy(kpat))) > amip) lmip = .true.
      enddo

      if (lmip) nmip = nmip + 1 
      lval(ipat) = (lval(ipat) .and. .not.lmip)  
!
! determine the orientation:
!
      if (lrec(ipat) .and. lval(ipat)) then
        k = npnt(ipat)
        if (k == 2) then
          if (i4h(3,2) /= 0) then
            itype = 1
          else if (i4h(2,1) /= 0) then
            itype = 2
          else if (i4h(1,2) /= 0) then
            itype = 3
          else if (i4h(2,3) /= 0) then
            itype = 4
          endif
        else if (k == 3 .or. k == 4) then
          if (i4h(3,2) /= 0) then
            if (i4h(2,1) /= 0) then
              itype = 1
            else
              itype = 4
            endif
          else if (i4h(1,2) /= 0) then
            if (i4h(2,1) /= 0) then
              itype = 2
            else
              itype = 3
            endif
          endif
        endif
      endif
!
! accumulate the distribution histogram: fill orientation IPAT3()
!
      if (lrec(ipat) .and. lval(ipat)) then
        do k = ind1(ipat), ind1(ipat) + npnt(ipat) - 1
          ix = ixx(k) - ix0
          iy = iyy(k) - iy0
          if (max(iabs(ix),iabs(iy)) <= 1) then
            if (i4map(ix+ix0,iy+iy0) > 0) &
                   r4buf(ix+2,iy+2) = r4buf(ix+2,iy+2) + 1.0
            ipat3(indmap(ixx(k),iyy(k))) = &
                   (ix+2)*100 + (iy+2)*10 + itype
          endif
        enddo
      endif

      if (lval(ipat)) then
        i = npnt(ipat)
      else
        i = -npnt(ipat)
      endif

      do k = ind1(ipat), ind1(ipat) + npnt(ipat) - 1
        ipat1(indmap(ixx(k),iyy(k))) = i
        ipat2(indmap(ixx(k),iyy(k))) = -ipat
        if (.not.lval(ipat)) then
!
! Create corresponding event quality flags:
! next to border of the CCD (or window)
!
          if (code == 1 ) iqual(indmap(ixx(k),iyy(k))) =   &
                  ior(iqual(indmap(ixx(k),iyy(k))),EVATT_CLOSE_TO_CCD_WINDOW) 

! next to a bright/hot pixel declared on-board
          if (code == 110) iqual(indmap(ixx(k),iyy(k))) =   &
                  ior(iqual(indmap(ixx(k),iyy(k))),EVATT_CLOSE_TO_ONBOARD_BADPIX)

! next to a bright/hot pixel not declared on-board
          if (code == 120 .or. code == 130) iqual(indmap(ixx(k),iyy(k))) =   &
                  ior(iqual(indmap(ixx(k),iyy(k))),EVATT_CLOSE_TO_BRIGHTPIX)

! next to a dead pixel      
          if (code == 310 .or. code == 320 .or. code == 330) &
             iqual(indmap(ixx(k),iyy(k))) =   &
                  ior(iqual(indmap(ixx(k),iyy(k))),EVATT_CLOSE_TO_DEADPIX)   

! "cosmic ray" event
          if (lmip) iqual(indmap(ixx(k),iyy(k))) =   &
                  ior(iqual(indmap(ixx(k),iyy(k))),EVATT_COSMIC_RAY)   

        endif 
      enddo
!
! fill pattern index IPAT2()
!
      if(lrec(ipat) .and. lval(ipat)) ipat2(indmap(ix0,iy0)) = ipat
!
! accumulate the distribution map:
!
      if (lrec(ipat) .and. lval(ipat)) then
        do iy = 1, 3
          do ix = 1, 3
            if (i4h(ix,iy) /= 0) i4dist(ix,iy)=i4dist(ix,iy)+1
          enddo
        enddo            
      endif
!
! only for high verbosities
!
      if (verbosity(AppMsg) >= Noisy) then
!
! fill r4map
!
        do i = 1, nevf
          if (lqual(i)) then
            ix = ixev(i) 
            iy = iyev(i)
            if (i2mapf(ix,iy) == 0) then
              r4map(ix,iy) = ampl(i)  
            endif
          endif
        enddo

        if (lrec(ipat) .and. lval(ipat)) then
          chbuf = "valid"
        else
          chbuf = "invalid"
        endif
        k = npnt(ipat)
        if (k == 2) then
          chbuf = chbuf(1:len(trim(chbuf))) // " double"   
        else if (k == 3) then
          chbuf = chbuf(1:len(trim(chbuf))) // " triple"
        else if (k == 4) then
          chbuf = chbuf(1:len(trim(chbuf))) // " quadruple"
        else
          chbuf = chbuf(1:len(trim(chbuf))) // " pattern"
        endif
        call message(AppMsg,Noisy,"")
        write(chbuf(len(trim(chbuf))+1:),"(i3,a)") ipat,":"
        write(errstr,"(a)") chbuf
        call message(AppMsg,Noisy)
        do iy = 3, 1, -1
          write(errstr,"(10x,3f8.2,5x,3i6)")                 &
                 (r4map(ix+ix0-2,iy+iy0-2),ix=1,3),  &
                 (nint(r4buf(ix,iy)),ix=1,3)
          call message(AppMsg,Noisy)
        enddo
!
!initialize r4map for next run
!
        do i = 1, nevf
          if (lqual(i)) then
            ix = ixev(i) 
            iy = iyev(i)
            if (i2mapf(ix,iy) == 0) then
              r4map(ix,iy) = 0.0  
            endif
          endif
        enddo

      endif

    enddo


!
! count the number of pattern sizes:
!

    n2 = 0
    n3 = 0
    n4 = 0
    nrest = 0
    do ipat = 1, npat
      k = npnt(ipat)
      if (lrec(ipat) .and. lval(ipat)) then
        if (k == 2) then
          n2 = n2 + 1
        else if (k == 3) then
          n3 = n3 + 1
        else if (k == 4) then
          n4 = n4 + 1
        endif
      else
        nrest = nrest + 1
      endif
    enddo
    !
    ! determine the largest pattern:
    !

    k = npnt(1)
    do ipat = 2, npat
      k = max(k,npnt(ipat))
    enddo
    nmax = k

  endif
        
!
! mark all single events at the edge of the CCD or touching bad
! pixels as invalid, and count the number of valid single events:
!

  n1 = 0
  code = 0
  do k = 1, nevf
    if(lqual(k)) then
      ix = ixev(k)
      iy = iyev(k)
      if (i4map(ix,iy) == -1) then
        if (i2mpcf(ix,iy) /= 0) then
          ipat1(indmap(ix,iy)) = -1
          code = i2mpcf(ix,iy)
!
! XMM: create corresponding single event quality flags:
!
! next to border of the CCD (or window)  
!
          if (code == 1 )           &
              iqual(indmap(ix,iy)) =   &
               ior(iqual(indmap(ix,iy)),EVATT_CLOSE_TO_CCD_WINDOW)
!
! next to a bright/hot pixel declared on-board
!
          if (code == 11)           &
              iqual(indmap(ix,iy)) =   &
               ior(iqual(indmap(ix,iy)),EVATT_CLOSE_TO_ONBOARD_BADPIX)
!
! next to a bright/hot pixel not declared on-board
!
          if (code == 12 .or. code == 13) &
              iqual(indmap(ix,iy)) =   &
               ior(iqual(indmap(ix,iy)),EVATT_CLOSE_TO_BRIGHTPIX) 
!     
! next to a dead pixel      
!
          if (code == 31 .or. code == 32 .or. code == 33) &
              iqual(indmap(ix,iy)) =   &
               ior(iqual(indmap(ix,iy)),EVATT_CLOSE_TO_DEADPIX)
        else
          ipat1(indmap(ix,iy)) = 1
          n1 = n1 + 1
        endif
      endif
    else
!
! flag pattern type as null
!
    ipat1(k) = 32767
    endif
  enddo
!
! initialize i4map for next run
!
  do k = 1, nevf
    if (lqual(k)) then
      ix = ixev(k) 
      iy = iyev(k)
      if (i2mapf(ix,iy) == 0) then
        i4map(ix,iy) = 0  
      endif
    endif
  enddo
!
! deallocate array dimensions
!
  deallocate(ieqv)
  deallocate(ieqvl)
  deallocate(ind1)
  deallocate(ind2)
  deallocate(ixx)
  deallocate(iyy)
  deallocate(npnt)
  deallocate(ncls)
  deallocate(inpat)
  deallocate(ixsel)
  deallocate(iysel)
  deallocate(izsel)
  deallocate(lind)
  deallocate(lrec)
  deallocate(lval)
  if (lscr) then
    deallocate(jpat1,jpat2,jpat3,jpat4)
  endif


end subroutine split


! ============================
! Collect Pattern Informations
! ============================
subroutine collectPattern(irow,irowen,nevf,ipat1,ipat2,ipat3,ipat4,iqual, &
!                         jqual,
                          lqual,i4buf,k1,k2,k3,k4,krest,kmax,kmip)

  integer,    intent(in)           :: irow, irowen, nevf, &
                                      i4buf(3,3), &
                                      k1, k2, k3, k4, krest, kmax, kmip
  integer(kind=Int32),dimension(:) :: iqual  ! ,jqual
  logical,dimension(:)             :: lqual
  integer(kind=Int16),dimension(:) :: ipat1,ipat2,ipat3,ipat4

  integer(kind=Int32)              :: i, ii
  integer(kind=int16)              :: i4dist(3,3), dfrom, dcode, &
                                      pat_ori_1, pat_ori_2
  integer(kind=int8)               :: from, code

  if (verbosity(AppMsg) >= Noisy) &
     call message(AppMsg,Noisy,"... entering collectPattern...")

!
! set invalid pattern information (e.g. trailing events) to NULL
!
  do i = 1, nevf 
    if (.not.lqual(i)) then
! ipat2 == pat_ind
      ipat2(i) = 32767
! ipat3 == pat_ori
      ipat3(i) = 32767
! ipat4 == pat_seq
      ipat4(i) = 255
    endif
  enddo

!
! append the pattern and flag information:
!
  pat_typCol(irow:irowen)  =  ipat1(1:nevf)
  pat_indCol(irow:irowen)  =  ipat2(1:nevf)
  pat_oriCol(irow:irowen)  =  ipat3(1:nevf)
  pat_seqCol(irow:irowen)  =  ipat4(1:nevf)
  qualflagCol(irow:irowen) =  iqual(1:nevf) 

!
! combine PAT_Infos into the PATTERN and PAT_ID columns:
!
  do i = irow, irowen
    ii = i - irow + 1

!
! initialize columns
!
    if (lqual(ii)) then
      patternCol(i) = 0
      pat_idCol(i)  = 0

!
! PATTERN ... the MOS code numbers 0-12 are covered 
!
      code = 0

      if (pat_typCol(i) < 0) then            ! invalid singles, doubles, etc.
        from = 1
        call mvbits(from,0,1,code,7)         ! set bit 8
      endif

      if (pat_indCol(i) < 0) then
        from = 1
        call mvbits(from,0,1,code,6)         ! set bit 7
      endif

      if (abs(pat_typCol(i)) == 1) then      ! single
        from = 0
        call mvbits(from,0,4,code,0)         ! unset bits 1-4
      elseif (abs(pat_typCol(i)) == 2) then      ! double
        from = 13                            ! 0->13 : 27-07-2000
        if (pat_oriCol(i) == 224 .or. &
            pat_oriCol(i) == 234) from = 1
        if (pat_oriCol(i) == 221 .or. &
            pat_oriCol(i) == 321) from = 2
        if (pat_oriCol(i) == 222 .or. &
            pat_oriCol(i) == 212) from = 3
        if (pat_oriCol(i) == 223 .or. &
            pat_oriCol(i) == 123) from = 4
        call mvbits(from,0,4,code,0)         ! unset bits 1-4 if invalid
      elseif (abs(pat_typCol(i)) == 3) then      ! triple
        from = 14                            ! 0->14 : 27-07-2000
        if (pat_oriCol(i) == 224 .or. &
            pat_oriCol(i) == 324 .or. &
            pat_oriCol(i) == 234) from = 5        
        if (pat_oriCol(i) == 221 .or. &
            pat_oriCol(i) == 211 .or. &
            pat_oriCol(i) == 321) from = 6
        if (pat_oriCol(i) == 222 .or. &
            pat_oriCol(i) == 122 .or. &
            pat_oriCol(i) == 212) from = 7
        if (pat_oriCol(i) == 223 .or. &
            pat_oriCol(i) == 233 .or. &
            pat_oriCol(i) == 123) from = 8
        call mvbits(from,0,4,code,0)         ! unset bits 1-4 if invalid
      elseif (abs(pat_typCol(i)) == 4) then      ! quadruple
        from = 15                            ! 0->15 : 27-07-2000
        if (pat_oriCol(i) == 224 .or. &
            pat_oriCol(i) == 324 .or. &
            pat_oriCol(i) == 334 .or. &
            pat_oriCol(i) == 234) from = 9
        if (pat_oriCol(i) == 221 .or. &
            pat_oriCol(i) == 211 .or. &
            pat_oriCol(i) == 311 .or. &
            pat_oriCol(i) == 321) from = 10
        if (pat_oriCol(i) == 222 .or. &
            pat_oriCol(i) == 122 .or. &
            pat_oriCol(i) == 112 .or. &
            pat_oriCol(i) == 212) from = 11
        if (pat_oriCol(i) == 223 .or. &
            pat_oriCol(i) == 233 .or. &
            pat_oriCol(i) == 133 .or. &
            pat_oriCol(i) == 123) from = 12
        call mvbits(from,0,4,code,0)         ! unset bits 1-4 if invalid
      elseif (abs(pat_typCol(i)) > 4) then       ! large pattern
                                             ! now use lowest 5 bits to store
                                             ! PAT_TYP - 5
        from = 1
        call mvbits(from,0,1,code,5)
        from = abs(pat_typCol(i)) - 5
        if (from > 36 - 5) from = 36 - 5     ! if PAT_TYP > 36 use 36 
        call mvbits(from,0,5,code,0)
      endif

      if (pat_typCol(i) > +36) code = 127    ! 27-07-2000 mjf
      if (pat_typCol(i) < -36) code = 255    ! 27-07-2000 mjf
      patternCol(i) = code

!
! PAT_ID
!
      dcode = 0
!
! PAT_IND
!
      dfrom = abs(pat_indCol(i))
      if (dfrom > 511) dfrom = 511           ! if PAT_IND > 511 use 511 
      call mvbits(dfrom,0,9,dcode,0)

!
! PAT_ORI
!
      if (pat_oriCol(i) /= 0) then
        pat_ori_1 = int(pat_oriCol(i) / 100)
        pat_ori_2 = int((pat_oriCol(i) - pat_ori_1 * 100) / 10)
        call mvbits(pat_ori_1,0,2,dcode,11)
        call mvbits(pat_ori_2,0,2,dcode,9)
      endif

      pat_idCol(i) = dcode

    else
      patternCol(i) = 13
      pat_idCol(i)  = 32767
    endif

  enddo     

  i4dist(1:3,1:3) = i4dist(1:3,1:3) + i4buf(1:3,1:3)

  n1 = n1 + k1
  n2 = n2 + k2
  n3 = n3 + k3
  n4 = n4 + k4
  nrest = nrest + krest
  nmax = max(kmax,nmax)
  nmip = nmip + kmip

end subroutine collectPattern


! =========================================
! Convert from raw- to detector coordinates 
! =========================================
subroutine raw2det(prandom,scale_x, scale_y)

  logical            ,intent(in)             :: prandom

  ! Units in arcsecs and offset to ensure positive coordinates

  real(kind=double),   parameter :: det_unit   = 0.05d0 ! arcsec per det_coord
  integer(kind=int16), parameter :: det_offset = 0
  integer(kind=int32), parameter :: detSysType = DETECTOR_OPTICSCENTER_SYSTEM
  !                               coordSysType = CAMCOORD2
  real(kind=double)              :: scale_x, scale_y

  integer(kind=int16),dimension(:), allocatable :: tempCol 
! real(kind=single),  dimension(:), pointer :: Xmm, Ymm, Zmm
  real(kind=double),  dimension(:), pointer :: Xmm, Ymm, Zmm

  real(kind=double)                        :: mm_per_pixel_x, mm_per_pixel_y, &
                                              plate_scale_x, plate_scale_y
  real(kind=double)                        :: step
  integer(kind=int16), dimension(4)        :: xcorn, ycorn
  integer(kind=int16)                    :: xmin, xmax, ymin, ymax
  integer                                  :: i
  logical(kind=boolean)                    :: Correct_CCF

  if (verbosity(AppMsg) >= Noisy) &  
    call message(AppMsg,Noisy,"... entering raw2det...")

  correct_CCF = .true.

  ! - - - - - -
  ! Pixel to mm
  ! - - - - - -
  call CAL_setState(EPN)

  if (correct_CCF) then

    call CAL_getMiscellaneousData("MM_PER_PIXEL_X",mm_per_pixel_x)
    call CAL_getMiscellaneousData("MM_PER_PIXEL_Y",mm_per_pixel_y)
    call CAL_getMiscellaneousData("PLATE_SCALE_X",plate_scale_x)
    call CAL_getMiscellaneousData("PLATE_SCALE_Y",plate_scale_y)

    scale_x = (plate_scale_x/mm_per_pixel_x) / det_unit
    scale_y = (plate_scale_y/mm_per_pixel_y) / det_unit

!   print*,'@@@@@@@@@@@@@@@@ CAL @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@'
!   print*,'MM_PER_PIXEL_X = ',MM_PER_PIXEL_X
!   print*,'MM_PER_PIXEL_Y = ',MM_PER_PIXEL_Y
!   print*,'PLATE_SCALE_X  = ',PLATE_SCALE_X
!   print*,'PLATE_SCALE_Y  = ',PLATE_SCALE_Y
!   print*,'DET_UNIT       = ',DET_UNIT
!   print*,'SCALE_X        = ',SCALE_X
!   print*,'SCALE_Y        = ',SCALE_Y
!   print*,'@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@'

  else

    scale_x = 27.5258  / det_unit  ! values supplied by DL 20000519_1142
    scale_y = 27.5258  / det_unit  ! values supplied by DL 20000519_1142

!   print*,'@@@@@@@@@@@@@ epevents @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@'
!   print*,'SCALE_X        = ',SCALE_X
!   print*,'SCALE_Y        = ',SCALE_Y
!   print*,'@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@'

  endif

  call setAttribute(events, "DET_SYST",detSysType, &
                    comment="System for camera coordinates")      

  call setAttribute(events, "SCALE_X",SCALE_X, &
                    comment="plate scale [(arcsec/mm)/(arcsec/sky_pix)]")      

  call setAttribute(events, "SCALE_Y",SCALE_Y, &
                    comment="plate scale [(arcsec/mm)/(arcsec/sky_pix)]")

  step = det_unit / 3600.d0 

  if (nevent > 0) then

    allocate(tempCol(0:nevent-1))

    if (linstmti .or. linstmbu) then     ! TIMING or BURST
      tempCol = srcpos
    else
      tempCol = rawYCol
    endif

    call CAL_setState(EPN,randomize = prandom)

    call CAL_rawXY2mm(rawXCol, tempCol, Xmm, Ymm, Zmm, detSysType)

    ! TEST: how does CAL_rawXY2mm perform the conversion?

!   if (verbosity(AppMsg) > Sparse) then
    if (verbosity(AppMsg) > Noisy) then
      write(errstr,*) &
             "Checking the CAL_rawXY2mm conversion with randomize =",prandom, &
             "  detSysType =",detSysType
      call message(AppMsg,Sparse)

      do i = 0, nevent - 1
        write(errstr,*) "i=",i,"rawX=",rawXCol(i),"rawY=",rawYCol(i), &
                        "Xmm=",Xmm(i),"Ymm=",Ymm(i),"Zmm=",Zmm(i)
        call message(AppMsg,Sparse)
      enddo
    endif  

    detXCol => int16Data(detX) 
    detYCol => int16Data(detY)
    
!   do i = 0, nevent-1
!     detXCol(i) = nint(Xmm(i) * scale_x)    
!     detYCol(i) = nint(Ymm(i) * scale_y)
!   enddo

    detXCol = nint(Xmm * scale_x)    
    detYCol = nint(Ymm * scale_y)

!   do i = 0, 20
!     write(*,'(5i8)') i,rawXCol(i),tempCol(i),detXCol(i),detYCol(i)
!   enddo

    deallocate(tempCol)      

    call CAL_releaseMemory(Xmm,Ymm,Zmm)

  endif

  ! ---------------------------------------------------------
  ! Write TLMIN/MAX and WCS keywords for detector coordinates
  ! The first part of TCTYP should not be RA, DEC
  ! ---------------------------------------------------------

  ! deduce extreme detector coordinates of current CCD

  xcorn = (/1,   1,  64, 64/)
  ycorn = (/2, 200, 200,  2/)

  call CAL_setState(EPN,randomize = .false.)
  call CAL_rawXY2mm(xcorn, ycorn, Xmm, Ymm, Zmm, detSysType)

  xcorn = nint(Xmm * scale_x)
  ycorn = nint(Ymm * scale_y)

  call CAL_releaseMemory(Xmm,Ymm,Zmm)

  xmin = minval(xcorn)
  xmax = maxval(xcorn)
  ymin = minval(ycorn)
  ymax = maxval(ycorn)

  call setAttribute(detX,"TCRPX",det_offset,comment="Center for DETX")
  call setAttribute(detX,"TCDLT",step,"degrees",  &
       comment="Pixel size for DETX")
  call setAttribute(detX,"TLMIN",xmin,comment="Minimum for DETX")
  call setAttribute(detX,"TLMAX",xmax,comment="Maximum for DETX")

  call setAttribute(detY,"TCRPX",det_offset,comment="Center for DETY")
  call setAttribute(detY,"TCDLT",step,"degrees",  &
       comment="Pixel size for DETY")
  call setAttribute(detY,"TLMIN",ymin,comment="Minimum for DETY")
  call setAttribute(detY,"TLMAX",ymax,comment="Maximum for DETY")

end subroutine raw2det  


! ======================
! Recombine split events 
! ======================
subroutine recombine(irow,irowen,nevf,ampl,ipat1,ipat2,iqual,lqual, &
                     lrecombraw,lflag1)

  ! the main pixels of unrecognized patterns will be assigned the
  ! negative sum of the corrected amplitudes of all pattern pixels

  integer,            intent(in)                :: irow, irowen, nevf
  real(kind=single),dimension(:)                :: ampl
  integer(kind=Int32),dimension(:)              :: iqual
  integer(kind=Int16),dimension(:)              :: ipat1, ipat2
  logical,dimension(:)                          :: lqual
  logical,            intent(in)                :: lrecombraw   
  logical,            intent(inout)             :: lflag1   

  integer             :: k, i, iamc(nevf), mr = 0, j, &
           iind(maxpat,maxcls), ipt, ipmx, &
                         num(maxpat)    ! number of splitpartners in pattern 
  real(kind=single)   :: sum(maxpat)
  logical(kind=bool)  :: lneg, patternNumberMissing


  if (lflag1) then
    lflag1 = .false.
    if (lrecombraw) then
      call message(AppMsg,Noisy, &
                  "... entering recombine: work on raw amplitudes only  ...")
    else
      call message(AppMsg,Noisy, &
                  "... entering recombine: work on calibrated amplitudes...")
    endif
  endif

  !
  ! get number of patterns
  !

  if (verbosity(AppMsg) > Noisy) call message(AppMsg,Noisy,' ')

  k = 0
  do i = 1, nevf
!   next line leads to failure for saslib-1.61 and later if old EVATTs are used
!   if (lqual(i)) k = max(k,int(abs(ipat2(i))))
    if (lqual(i)) then
      if (ipat2(i) < 32767) then
        k = max(k,int(abs(ipat2(i))))
      else
        write(errstr,*) 'RECOMBINE(): unusual value LQUAL(I)=T: ', &
                      'NEVF,I,K,IPAT1(I),IPAT2(I) = ', &
                       NEVF,I,K,IPAT1(I),IPAT2(I)
        call warning('unusualQual')
        lqual(i) = .false.
      endif
    endif
    if (verbosity(AppMsg) > Noisy) then
      write(errstr,*) 'RECOMBINE(): NEVF,I,LQUAL(I),K,IPAT1(I),IPAT2(I) = ', &
                       NEVF,I,LQUAL(I),K,IPAT1(I),IPAT2(I)
      call message(AppMsg,Noisy)
    endif
  enddo
  npat = k

  if (verbosity(AppMsg) > Noisy) call message(AppMsg,Noisy,' ')

!
! determine sums of all partners of individual patterns
!

  if (verbosity(AppMsg) > Noisy) then
    write(errstr,*) 'RECOMBINE(): MAXPAT,NPAT = ',MAXPAT,NPAT
    call message(AppMsg,Noisy)
  endif

  sum(1:npat) = 0.0

  do i = 1, nevf
    if(lqual(i)) then
      k = ipat2(i)
      if (k /= 0) then
        if (k < 0) k = - k
        sum(k) = sum(k) + nint(ampl(i))
        if (sum(k) >= 32767) then
          sum(k) = 32767
        endif
      endif
    endif
  enddo

  !
  ! assignment of combined amplitudes to events
  !

  do i = 1, nevf
    if(lqual(i)) then
      k = ipat2(i)
      if (lrecombraw) then
        if (k < 0) k = - k
        !
        ! for gain correction input from raw amplitudes
        !
        ! accept all singles, valid patterns, invalid doubles
        ! and invalid doubles

        iamc(i) = 0

        ! valid patterns:

        if (ipat1(i) > 1 .and. ipat1(i) < 5) iamc(i) = nint(sum(k))

        ! singles:

        if (abs(ipat1(i)) == 1)              iamc(i) = nint(ampl(i))

        ! invalid doubles:

        if (ipat1(i) == -2)                  iamc(i) = nint(sum(k))

      else

        !
        ! for recombination of calibrated amplitudes in recognized patterns
        !    

        if (k < 0) then                 ! not main event
          iamc(i) = 0
        else if (k == 0) then           ! valid singles
          iamc(i) = nint(ampl(i))
        else                            ! main event
          iamc(i) = nint(sum(k))
        endif

      endif
    endif
  enddo

  if (.not.lrecombraw) then ! do full run  

    ! assign the sum of the corrected amplitudes of an unrecognized
    ! pattern to the pixel with the highest charge:

    !
    ! define number of split partners in pattern (num(i)) and
    ! fill array (iind) with adress of split partners in pattern 
    !

    num(1:npat) = 0

    do i = 1, npat
      do k = 1, nevf
        if(lqual(k)) then
          if (ipat2(k) == i .or. ipat2(k) == - i) then
            num(i) = num(i) + 1
            if (num(i) > maxcls) then
              ! maximum allowed number of entries in event class is exceeded
              ! increase parameter maxcls
              write(errstr,"(a,2i8)") &
                   "Frame counter, pattern number:", ifrm, i
              call message(AppMsg,Sparse)
              write(errstr,"(a,i8)") &
                   "Maximum number of entries in event class exceeded:", maxcls
              call message(AppMsg,Sparse)
              call message(AppMsg,Sparse,"increase parameter maxcls")
              call deleteDataSet(outfilename)
              call error("tooManyClassentries", &
                  "Number of entries in event class exceeds the limit")
            endif
            iind(i,num(i)) = k
          endif
        endif
      enddo
    enddo

    !
    ! loop over all patterns
    !

    do i = 1, npat

      ! find unrecognizable patterns:
      ! LNEG = TRUE for all events of pattern or NUM(I) < 1

      lneg = .true.
      do k = 1, num(i)
        lneg = (lneg .and. ipat2(iind(i,k)) < 0)
      enddo

      ! for unrecognizable patterns determine maximum index and assign sum

      patternNumberMissing = .false.
      if (lneg) then
        ipt = 1
        mr = irow+iind(i,ipt)-1
        if (mr < 0) then
          write(errstr,*) 'Array index of unrecognizable pattern inconsistent.'//&
                         ' This is uncritical if you have set withoutoftime=Y'
          call message(AppMsg,Sparse)
          write(errstr,*) 'Values i,ipt,irow,iind(i,ipt),num(i),npat,nevf =',&
                                  i,ipt,irow,iind(i,ipt),num(i),npat,nevf
          call message(AppMsg,Sparse)
          call warning('unrecognizedPatternIndex',' ')
          write(errstr,*) '   k lqual(k) iqual(k) ipat1(k) ipat2(k) ampl(k) iamc(k):'
          call message(AppMsg,Sparse)
          do k = 1,nevf
            write(errstr,'(i5,l8,3i9,2i8)') & 
                     k,lqual(k),iqual(k),ipat1(k),ipat2(k),nint(ampl(k)),iamc(k)
            call message(AppMsg,Sparse)
          enddo
          do k = 1,npat
            write(errstr,'(i5,20i5)') k,num(k),(iind(k,j),j=1,npat)
            call message(AppMsg,Sparse)
          enddo
          patternNumberMissing = .true.
        else
          ipmx = pha_ctiCol(mr)
        endif
        do k = 2, num(i)
          mr = irow+iind(i,k)-1
          if (mr < 0) then
            write(errstr,*) 'Values i,k,irow,iind(i,k),num(i),npat,nevf =',&
                                    i,k,irow,iind(i,k),num(i),npat,nevf
            call warning('unrecognizedPatternIndex')
          else
            if (pha_ctiCol(mr) > ipmx) then
              ipmx = pha_ctiCol(mr)
              ipt = k
            endif
          endif
        enddo
!       if (.not. patternNumberMissing) &
        iamc(iind(i,ipt)) = -nint(sum(i))
      endif

    enddo

    do  i = 1, nevf
!
! XMM SAS: change PI to -PI
! for invalid singles:
!
      if (ipat1(i) == -1)  iamc(i) = -iamc(i)

! Change too big recombined amplitude to NullValue
      if (iamc(i) <= -32767) then
        call setToNull(pi,irow-1+i)
      else if (.not.lqual(i)) then
        call setToNull(pi,irow-1+i)
      else if (iamc(i) == 0) then
        call setToNull(pi,irow-1+i)
      else
        piCol(irow-1+i) = iamc(i)
      endif

! Flag for "secondary" pixel in an event
      if (ipat2(i) < 0 .and. iamc(i) == 0) &
        iqual(i) = ior(iqual(i),EVATT_SECONDARY)

    enddo

    qualflagCol(irow:irowen) = iqual(1:nevf)

  endif

end subroutine recombine


! ====================================
! Add attributes to EXPOSURE extension
! ====================================
subroutine addExposure(ithp,prandom,erandom,intemperature,gainctiaccuracy, &
                       scale_x, scale_y, twidth, corrpar, withphotonmap)

  real(kind=double)   ,intent(in)   :: intemperature, scale_x, scale_y, &
                                       corrpar(12)
  integer(kind=Int32) ,intent(in)   :: gainctiaccuracy
  integer(kind=Int16) ,intent(in)   :: ithp
  logical             ,intent(in)   :: prandom, erandom, twidth, withphotonmap
  character(len=8)                  :: extname

  call message(AppMsg,Noisy,"... entering addExposure...")

  write(extname,'(a)') "EXPOSURE"
  exposure = Table(outfile,extname,MODIFY)

  call setAttribute(exposure, "SCALE_X",SCALE_X, &
                    comment="plate scale [(arcsec/mm)/(arcsec/sky_pix)]")      

  call setAttribute(exposure, "SCALE_Y",SCALE_Y, &
                    comment="plate scale [(arcsec/mm)/(arcsec/sky_pix)]")

  ! -------------------------------------
  ! Add epevents parameters as attributes
  ! -------------------------------------
  call setAttribute(exposure, "REEM_THR",ithp, & 
       comment="minimum PHA of preceeding event for reemission check")
  call setAttribute(exposure, "RANDOM_P",prandom, "randomize pixel position")
  call setAttribute(exposure, "RANDOM_E",erandom, "randomize amplitudes in bin")
  call setAttribute(exposure, "TWIDTH",twidth, "non-standard energy bin width")
  call setAttribute(exposure, "CCD_TEMP",intemperature,"K", &
       comment="CCD temperature")
  call setAttribute(exposure, "CTI_ACCU",gainctiaccuracy, &
       comment="accuracy of gain/cti correction")

  ! ------------------------------------
  ! Add pattern statistics as attributes
  ! ------------------------------------

  call setAttribute(exposure,"ANALYSED",nevent,"", &
       "number of analysed events")
  call setAttribute(exposure,"SINGLES" ,n1,   "", &
       "number of single events")
  call setAttribute(exposure,"DOUBLES" ,n2,   "", &
       "number of double events")
  call setAttribute(exposure,"TRIPLES" ,n3,   "", &
       "number of triple events")
  call setAttribute(exposure,"QUADRUPL",n4,   "", &
       "number of quadruple events")
  call setAttribute(exposure,"NOTRECEV",n0,   "", &
       "number of not recognized events")
  call setAttribute(exposure,"NOTRECPA",nrest,"", &
       "number of not recognized patterns")
  call setAttribute(exposure,"MAXPAT"  ,nmax, "", &
       "maximum pattern size")
  call setAttribute(exposure,"TRAILING"    ,ntrail, "", &
       "number of trailing events")
  call setAttribute(exposure,"MIPS"    ,nmip, "", &
       "number of MIPs found")
  call setAttribute(exposure,"RECPHOTO",n1+n2+n3+n4,"", &
       "number of recognized photons")
  call setAttribute(exposure, "SCR_88",enable_scr_88, &
        comment="use PATTERN coding as of SOC_SCR_88")


  ! ------------------------------------
  ! Add additional correction parameters
  ! ------------------------------------

  call setAttribute(exposure,"ADDCORRB","------------------",&
                 "- Begin of list of additional corrections -----")
  call setAttribute(exposure,"GAINSCAL",corrpar(1),"",&
        comment="gain scaling Orsay/in-orbit")
  call setAttribute(exposure,"TYR_2000",corrpar(2),"yr",&
        comment="years elapsed since 1.1.2000")
  call setAttribute(exposure,"ADU_REF", corrpar(3),"adu",&
        comment="reference channel used for gain/CTI corrections")
  call setAttribute(exposure,"LTINCCTI",corrpar(4),"adu/yr",&
        comment="effect of long-term CTI increase")
  call setAttribute(exposure,"CCD_OFFS",corrpar(5),"adu",&
        comment="offset of individual CCD")
  call setAttribute(exposure,"TG_OFFS", corrpar(6),"adu",&
        comment="temperature/gain-relation offset")
  call setAttribute(exposure,"TG_SLOPE",corrpar(7),"adu/K",&
        comment="temperature/gain-relation slope")
  call setAttribute(exposure,"TG_TEMP", corrpar(8),"deg C",&
        comment="temperature/gain-relation temperature")
  call setAttribute(exposure,"LT_CTI_0", corrpar(9),"",&
        comment="CTI increase energy function: term order 0")
  call setAttribute(exposure,"LT_CTI_1", corrpar(10),"",&
        comment="CTI increase energy function: term order 1")
  call setAttribute(exposure,"LT_CTI_2", corrpar(11),"",&
        comment="CTI increase energy function: term order 2")
  call setAttribute(exposure,"LT_CTI_Y", corrpar(12),"pix",&
        comment="CTI increase reference Y position")
  call setAttribute(exposure,"ADDCORRE","------------------",&
                 "- End of list of additional corrections -------")

  ! ----------------------------------------------------------------
  ! evlistcomb does not recognize BURST mode, change BURST to TIMING
  ! ----------------------------------------------------------------

  if (lburst) call setattribute(exposure,"DATAMODE","TIMING", &
                        "Instrument mode (changed from BURST to TIMING)")
!-----
! if a photon map is created then keep the EXPOSURE extension open to access keywords
!-----

  if (.NOT. withphotonmap) call release(exposure)

end subroutine addExposure



! ============================================================================
! make photon map title description
! ============================================================================
subroutine makeMapTitle(lothresh,hithresh,mapPatternType,map_title)

  real(kind=single), intent(in)            :: lothresh,hithresh
  character(len=*),  intent(in)            :: mapPatternType
  character(len=*),  intent(out)           :: map_title
  real(kind=single)                        :: enl,enh
  character(len=80)                        :: patTypeText

  enl = NINT(lothresh*1.e+2)*1.e-5
  enh = NINT(hithresh*1.e+2)*1.e-5

  select case (mapPatternType)
    case ("s")
      patTypeText = "Singles"
    case ("d")
      patTypeText = "Doubles"
    case ("t")
      patTypeText = "Triples"
    case ("q")
      patTypeText = "Quadruples"
    case ("n")
      patTypeText = "Non-singles"
    case ("u")
      patTypeText = "Unrecognized events"
    case ("r")
      patTypeText = "Recognized events"
    case ("m")
      patTypeText = "Multiple events"
    case ("e")
      patTypeText = "Re-emission events"
    case ("x")
      patTypeText = "Doubles in x-direction"
    case ("y")
      patTypeText = "Doubles in y-direction"
    case ("v")
      patTypeText = "Doubles in Y with main toward CAMEX"
    case ("w")
      patTypeText = "Doubles in Y with main away CAMEX"
    case default
      patTypeText = "Unknown type"
  end select

  write(map_title,'(a,a,f7.2,a,f7.2,a)') &
      trim(patTypetext),', $',enl,' - ',enh,'$\,keV'

end subroutine makeMapTitle


! ============================================================================
! fill photon maps depending on pattern types and energy ranges
! ============================================================================
subroutine fillPhotonMap(b1_array,b2_array,b3_array,b4_array,nx_ccd,ny_ccd, &
                         lothresh,hithresh,mapPatternType,nphotonband)

  integer, intent(in)                     :: nx_ccd, ny_ccd, nphotonband
  integer(kind=Int32), intent(out)        :: b1_array(1:nx_ccd,1:ny_ccd), &
                                             b2_array(1:nx_ccd,1:ny_ccd), &
                                             b3_array(1:nx_ccd,1:ny_ccd), &
                                             b4_array(1:nx_ccd,1:ny_ccd)
  real(kind=single), intent(in)           :: lothresh(1:nphotonband), &
                                             hithresh(1:nphotonband)
  character(len=1), intent(in)            :: mapPatternType(1:nphotonband)
  logical(kind=bool)                      :: is_single, is_double, is_triple,&
                                             is_quadruple, is_nonsingle, &
                                             is_unrecognized, is_all, &
                                             is_recognized, is_multiple, &
                                             is_double_x, is_double_y, &
                                             is_double_y_main_near_camex, &
                                             is_double_y_main_away_camex
  logical(kind=bool)                      :: lb(1:10), lp(1:10)
  integer(kind=Int32)                     :: ir, ib, ix, iy

  call message(AppMsg,Noisy,"... entering fillPhotonMap...")

  do ir = 0, nevent-1
    is_single =  (patternCol(ir) == 0)
    is_double = ((patternCol(ir) >= 1) .and. (patternCol(ir) <= 4))
    is_triple = ((patternCol(ir) >= 5) .and. (patternCol(ir) <= 8))
    is_quadruple = ((patternCol(ir) >= 9) .and. (patternCol(ir) <= 12))
    is_nonsingle = ((.NOT.is_single) .and. (ABS(pat_typCol(ir)) > 1 ))
    is_recognized = ((patternCol(ir) >= 0) .and. (patternCol(ir) <= 12))
    is_unrecognized = (.NOT.is_recognized)
    is_multiple = (is_recognized .AND. is_nonsingle)
    is_all = .true.
    is_double_x = ((patternCol(ir) == 2) .or. (patternCol(ir) == 4))
    is_double_y_main_near_camex = (patternCol(ir) == 1)
    is_double_y_main_away_camex = (patternCol(ir) == 3)
    is_double_y = (is_double_y_main_near_camex .or. is_double_y_main_away_camex)
    do ib = 1,nphotonband
      lb(ib) =  (piCol(ir) >= lothresh(ib) .and. piCol(ir) < hithresh(ib))
    enddo
    do ib = 1,nphotonband
      select case (mapPatternType(ib))
        case ("s")
          lp(ib) = is_single
        case ("d")
          lp(ib) = is_double
        case ("t")
          lp(ib) = is_triple
        case ("q")
          lp(ib) = is_quadruple
        case ("n")
          lp(ib) = is_nonsingle
        case ("u")
          lp(ib) = is_unrecognized
        case ("r")
          lp(ib) = is_recognized
        case ("m")
          lp(ib) = is_multiple
        case ("x")
          lp(ib) = is_double_x
        case ("y")
          lp(ib) = is_double_y
        case ("v")
          lp(ib) = is_double_y_main_near_camex
        case ("w")
          lp(ib) = is_double_y_main_away_camex
        case default
          lp(ib) = .false.
      end select
    enddo
    ix = rawXCol(ir)
    iy = rawYCol(ir)
!   do ib = 1,nphotonband
      if (lb(1) .and. lp(1)) b1_array(ix,iy) = b1_array(ix,iy)+1
      if (lb(2) .and. lp(2)) b2_array(ix,iy) = b2_array(ix,iy)+1
      if (lb(3) .and. lp(3)) b3_array(ix,iy) = b3_array(ix,iy)+1
      if (lb(4) .and. lp(4)) b4_array(ix,iy) = b4_array(ix,iy)+1
!   enddo

  enddo

  call message(AppMsg,Noisy,"... leaving fillPhotonMap...")

  end subroutine fillPhotonMap



! ===========================
! Produce equivalence classes
! ===========================
subroutine spleqv(neqv,ieqv,neqvl,ieqvl,ncls)

  ! neqv                  : number of equivalence pointers
  ! ieqv(maxeqv,2)        : equivalence pointers
  ! neqvl                 : number of equivalence classes
  ! ieqvl(maxeqv,maxcls)  : equivalence classes
  ! ncls(maxeqv)          : number of entries in each equivalence class

  implicit none

  integer, intent(in)    :: neqv
  integer, intent(out)   :: neqvl 
  integer,dimension(:,:) :: ieqv,ieqvl
  integer,dimension(:)   :: ncls
  integer                :: k, k1, k2, &
                            kn1, kn2, &
                            keqv, keqvl, keqvl1, keqvl2, &
                            kcls, kcls1, kcls2, kval, kk
  logical                :: lfound, lnew

  if (verbosity(AppMsg) >= Noisy) &
      call message(AppMsg,Noisy,"... entering spleqv...")

  neqvl = 0
  do keqv = 1, neqv
    k1 = ieqv(keqv,1)
    k2 = ieqv(keqv,2)
    keqvl = 0
    lfound = .false.
    do while (keqvl < neqvl .and. .not.lfound)
      keqvl = keqvl + 1
      do kcls = 1, ncls(keqvl)
        do kn1 = k1, k2, k2 - k1
          if (kn1 == k1) then
            kn2 = k2
          else
            kn2 = k1
          endif
          if (kn1 == ieqvl(keqvl,kcls)) then
            k = 0
            lnew = .true.
            do while (k < ncls(keqvl) .and. lnew)
              k = k + 1
              lnew = (kn2 /= ieqvl(keqvl,k))
            enddo
            if (lnew) then
              ncls(keqvl) = ncls(keqvl) + 1
              if (ncls(keqvl) > maxcls) then
                ! maximum allowed number of entries in event class is exceeded
                ! increase parameter maxcls
                write(errstr,"(a,i8)") &
                 "Maximum number of entries in event class exceeded:", maxcls
                call message(AppMsg,Sparse)
                call message(AppMsg,Sparse,"increase parameter maxcls")
                call deleteDataSet(outfilename)
                call error("tooManyClassentries", &
                  "Number of entries in event class exceeds the limit")
              endif
              ieqvl(keqvl,ncls(keqvl)) = kn2
            endif
            lfound = .true.
          endif
        enddo
      enddo
    enddo
    if (.not.lfound) then
      neqvl = neqvl + 1
      ieqvl(neqvl,1) = k1
      ieqvl(neqvl,2) = k2
      ncls(neqvl) = 2
    endif
  enddo

  !
  ! combine equivalence classes, if necessary:
  !

  do keqvl1 = 1, neqvl
    kcls1 = 0
    do while (kcls1 < ncls(keqvl1))
      kcls1 = kcls1 + 1
      k1 = ieqvl(keqvl1,kcls1)
      do keqvl2 = keqvl1 + 1, neqvl
        kcls2 = 0
        do while (kcls2 < ncls(keqvl2))
          kcls2 = kcls2 + 1
          k2 = ieqvl(keqvl2,kcls2)
          if (k1 == k2) then
            k = ncls(keqvl1)
            kcls = 0
            do while (kcls < ncls(keqvl2))
              kcls = kcls + 1
              kval = ieqvl(keqvl2,kcls)
              lnew = .true.
              do kk = 1, ncls(keqvl1)
                lnew = (lnew .and. (ieqvl(keqvl1,kk) /= kval))
              enddo
              if (lnew) then
                k = k + 1
                ieqvl(keqvl1,k) = kval
              endif
            enddo
            ncls(keqvl1) = k
            ncls(keqvl2) = 0
          endif
        enddo
      enddo
    enddo
  enddo

  !
  ! put the smallest entry at first position:
  !

  do keqvl = 1, neqvl
    k1 = ieqvl(keqvl,1)
    do kcls = 2, ncls(keqvl)
      k2 = ieqvl(keqvl,kcls)
      if (k2 < k1) then
        k1 = k2
        k = kcls
      endif
    enddo
    k2 = ieqvl(keqvl,1)
    ieqvl(keqvl,1) = k1
    ieqvl(keqvl,k) = k2
  enddo

end subroutine spleqv


! ===========================
! Print 1 frame of CCD events
! ===========================
subroutine print_CCD_ascii(ct)

  implicit none

  character(len=1),intent(in) :: ct(1:EPN_MAX_X,1:EPN_MAX_Y)
  integer                     :: i,j

  if (verbosity(AppMsg) >= Noisy) &
      call message(AppMsg,Noisy,"... entering print_CCD_ascii...")

  call message(AppMsg,Sparse,' ')
  write(errstr,'(1x,66a1)') "/",("-",j=1,EPN_MAX_X),"\"
  call message(AppMsg,Sparse)
  do i = EPN_MAX_Y,1,-1
    write(errstr,'(1x,66a1,i4)') '|',(ct(j,i),j=1,64),'|',i
    call message(AppMsg,Sparse)
  enddo
  write(errstr,'(1x,66a1)') "\",("-",j=1,EPN_MAX_X),"/"
  call message(AppMsg,Sparse)
  call message(AppMsg,Sparse,' ')

end subroutine print_CCD_ascii

end module epevents_mod
